"""WHAD Protocol BLE pdu messages abstraction layer.
"""
import struct
from scapy.compat import raw
from scapy.layers.bluetooth4LE import BTLE, BTLE_DATA, BTLE_CTRL, BTLE_ADV, BTLE_ADV_IND, \
    BTLE_ADV_NONCONN_IND, BTLE_ADV_DIRECT_IND, BTLE_ADV_SCAN_IND, BTLE_SCAN_RSP, BTLE_RF
from whad.hub.message import AbstractPacket, pb_bind, PbFieldInt, PbFieldBytes, PbMessageWrapper, \
    PbFieldBool, dissect_failsafe
from whad.hub.ble import Direction, AdvType, AddressType, BDAddress, BleDomain, BLEMetadata

from struct import pack

# correlation table
SCAPY_CORR_ADV = {
    AdvType.ADV_IND: BTLE_ADV_IND,
    AdvType.ADV_NONCONN_IND: BTLE_ADV_NONCONN_IND,
    AdvType.ADV_DIRECT_IND: BTLE_ADV_DIRECT_IND,
    AdvType.ADV_SCAN_IND: BTLE_ADV_SCAN_IND,
    AdvType.ADV_SCAN_RSP: BTLE_SCAN_RSP
}


def packet_to_bytes(packet):
    """Convert packet to bytes
    """
    try:
        return raw(packet)
    except TypeError:
        return bytes(packet.__bytes__())

# correlation table
SCAPY_CORR_ADV = {
    AdvType.ADV_IND: BTLE_ADV_IND,
    AdvType.ADV_NONCONN_IND: BTLE_ADV_NONCONN_IND,
    AdvType.ADV_DIRECT_IND: BTLE_ADV_DIRECT_IND,
    AdvType.ADV_SCAN_IND: BTLE_ADV_SCAN_IND,
    AdvType.ADV_SCAN_RSP: BTLE_SCAN_RSP
}

# correlation table
SCAPY_CORR_ADV_INV = {
    BTLE_ADV_IND: AdvType.ADV_IND,
    BTLE_ADV_NONCONN_IND: AdvType.ADV_NONCONN_IND,
    BTLE_ADV_DIRECT_IND: AdvType.ADV_DIRECT_IND,
    BTLE_ADV_SCAN_IND: AdvType.ADV_SCAN_IND,
    BTLE_SCAN_RSP: AdvType.ADV_SCAN_RSP
}

@pb_bind(BleDomain, "set_adv_data", 1)
class SetAdvData(PbMessageWrapper):
    """BLE set advertising data message class
    """
    scan_data = PbFieldBytes("ble.set_adv_data.scan_data")
    scanrsp_data = PbFieldBytes("ble.set_adv_data.scanrsp_data")

@pb_bind(BleDomain, "send_raw_pdu", 1)
class SendBleRawPdu(PbMessageWrapper):
    """BLE send raw PDU message class
    """
    direction = PbFieldInt("ble.send_raw_pdu.direction")
    conn_handle = PbFieldInt("ble.send_raw_pdu.conn_handle")
    access_address = PbFieldInt("ble.send_raw_pdu.access_address")
    pdu = PbFieldBytes("ble.send_raw_pdu.pdu")
    crc = PbFieldInt("ble.send_raw_pdu.crc")
    encrypt = PbFieldBool("ble.send_raw_pdu.encrypt")

    @dissect_failsafe
    def to_packet(self):
        """Convert message to the corresponding Scapy packet
        """
        print(self)
        packet = BTLE(access_addr=self.access_address, crc=self.crc)/self.pdu

        # Set packet metadata
        packet.metadata = BLEMetadata()
        packet.metadata.connection_handle = self.conn_handle
        packet.metadata.encrypt = self.encrypt
        packet.metadata.direction = self.direction
        packet.metadata.raw = True

        return packet

    @staticmethod
    def from_packet(packet, encrypt=False):
        """Convert packet to SendBlePdu message.
        """
        direction = packet.metadata.direction
        connection_handle = packet.metadata.connection_handle

        # Extract PDU
        if BTLE_DATA in packet:
            pdu = raw(packet[BTLE_DATA:])
        elif BTLE_CTRL in packet:
            pdu = raw(packet[BTLE_CTRL:])
        elif BTLE_ADV in packet:
            pdu = raw(packet[BTLE_ADV:])
        else:
            return None

        return SendBleRawPdu(
            direction=direction,
            pdu=pdu,
            conn_handle=connection_handle,
            access_address=BTLE(raw(packet)).access_addr,
            crc=BTLE(raw(packet)).crc,
            encrypt=encrypt
        )

@pb_bind(BleDomain, "send_pdu", 1)
class SendBlePdu(PbMessageWrapper):
    """BLE send PDU message class
    """
    direction = PbFieldInt("ble.send_pdu.direction")
    conn_handle = PbFieldInt("ble.send_pdu.conn_handle")
    pdu = PbFieldBytes("ble.send_pdu.pdu")
    encrypt = PbFieldBool("ble.send_pdu.encrypt")

    @dissect_failsafe
    def to_packet(self):
        """Convert message to the corresponding Scapy packet
        """
        packet = BTLE_DATA(self.pdu)

        # Set packet metadata
        packet.metadata = BLEMetadata()
        packet.metadata.connection_handle = self.conn_handle
        packet.metadata.encrypt = self.encrypt
        packet.metadata.direction = self.direction
        packet.metadata.raw = False

        return packet 

    @staticmethod
    def from_packet(packet, encrypt=False):
        """Convert packet to SendBlePdu message.
        """
        direction = packet.metadata.direction
        connection_handle = packet.metadata.connection_handle

        # Extract PDU
        if BTLE_DATA in packet:
            pdu = packet_to_bytes(packet[BTLE_DATA:])
        elif BTLE_CTRL in packet:
            pdu = packet_to_bytes(packet[BTLE_CTRL:])
        elif BTLE_ADV in packet:
            pdu = packet_to_bytes(packet[BTLE_ADV:])
        else:
            return None

        # Create a SendPdu message
        return SendBlePdu(
            direction=direction,
            conn_handle=connection_handle,
            pdu=pdu,
            encrypt=encrypt
        )



@pb_bind(BleDomain, "adv_pdu", 1)
class BleAdvPduReceived(PbMessageWrapper):
    """BLE advertising PDU received message class
    """
    adv_type = PbFieldInt("ble.adv_pdu.adv_type")
    rssi = PbFieldInt("ble.adv_pdu.rssi")
    bd_address = PbFieldBytes("ble.adv_pdu.bd_address")
    adv_data = PbFieldBytes("ble.adv_pdu.adv_data")
    addr_type = PbFieldInt("ble.adv_pdu.addr_type")

    @dissect_failsafe
    def to_packet(self):
        """Convert message into its corresponding Scapy packet
        """
        if self.adv_type in SCAPY_CORR_ADV:
            data = bytes(self.adv_data)

            packet = BTLE_ADV()/SCAPY_CORR_ADV[self.adv_type](
                    bytes(self.bd_address) + data
                )

            # Set TxAdd to 1 if address is random
            if self.addr_type == AddressType.RANDOM:
                packet.TxAdd = 1

            # Set packet metadata
            packet.metadata = BLEMetadata()
            packet.metadata.direction = Direction.UNKNOWN
            packet.metadata.rssi = self.rssi
            packet.metadata.raw = False

            # Success, return Scapy packet
            return packet
        else:
            # Unkown advertisement type
            return None

    @staticmethod
    def from_packet(packet):
        """Convert packet into BleAdvPduReceived message
        """
        if BTLE_ADV in packet:
            # Search advertisement type
            for adv_class in SCAPY_CORR_ADV_INV:
                if  packet.haslayer(adv_class):
                    adv_data = b''.join([bytes(x) for x in packet.getlayer(adv_class).data])
                    return BleAdvPduReceived(
                        adv_type=SCAPY_CORR_ADV_INV[adv_class],
                        rssi=packet.metadata.rssi if packet.metadata is not None else 0,
                        bd_address=BDAddress(packet.AdvA).value,
                        adv_data=adv_data,
                        addr_type=AddressType.RANDOM if packet.getlayer(BTLE_ADV).TxAdd == 1 else AddressType.PUBLIC
                    )
        else:
            # Error
            return None


@pb_bind(BleDomain, "pdu", 1)
class BlePduReceived(PbMessageWrapper):
    """BLE PDU received message class
    """
    direction = PbFieldInt("ble.pdu.direction")
    pdu = PbFieldBytes("ble.pdu.pdu")
    conn_handle = PbFieldInt("ble.pdu.conn_handle")
    processed = PbFieldBool("ble.pdu.processed")
    decrypted = PbFieldBool("ble.pdu.decrypted")

    @dissect_failsafe
    def to_packet(self):
        """Convert message into its corresponding Scapy packet
        """
        packet = BTLE_DATA(self.pdu)
        packet.metadata = BLEMetadata()
        packet.metadata.connection_handle = self.conn_handle
        packet.metadata.direction = self.direction
        packet.metadata.decrypted = self.decrypted
        packet.metadata.processed = self.processed
        packet.metadata.raw = False
        return packet

    @staticmethod
    def from_packet(packet):
        """Convert packet into BlePduReceived message
        """
        return BlePduReceived(
            pdu=bytes(packet),
            direction=packet.metadata.direction,
            conn_handle=packet.metadata.connection_handle,
            processed=packet.metadata.processed,
            decrypted=packet.metadata.decrypted
        )


@pb_bind(BleDomain, "raw_pdu", 1)
class BleRawPduReceived(PbMessageWrapper):
    """BLE raw PDU received message class
    """
    direction = PbFieldInt("ble.raw_pdu.direction")
    channel = PbFieldInt("ble.raw_pdu.channel")
    rssi = PbFieldInt("ble.raw_pdu.rssi")
    timestamp = PbFieldInt("ble.raw_pdu.timestamp")
    relative_timestamp = PbFieldInt("ble.raw_pdu.relative_timestamp")
    crc_validity = PbFieldBool("ble.raw_pdu.crc_validity")
    access_address = PbFieldInt("ble.raw_pdu.access_address")
    pdu = PbFieldBytes("ble.raw_pdu.pdu")
    crc = PbFieldInt("ble.raw_pdu.crc")
    conn_handle = PbFieldInt("ble.raw_pdu.conn_handle")
    processed = PbFieldBool("ble.raw_pdu.processed")
    decrypted = PbFieldBool("ble.raw_pdu.decrypted")

    @dissect_failsafe
    def to_packet(self):
        """Convert message into its corresponding Scapy packet
        """
        packet = BTLE(bytes(struct.pack("I", self.access_address)) + bytes(self.pdu) + bytes(struct.pack(">I", self.crc)[1:]))

        # Populate metadata
        packet.metadata = BLEMetadata()
        packet.metadata.direction = self.direction
        packet.metadata.connection_handle = self.conn_handle
        packet.metadata.channel = self.channel
        packet.metadata.processed = self.processed
        packet.metadata.raw = True

        if self.rssi is not None:
            packet.metadata.rssi = self.rssi
        if self.timestamp is not None:
            packet.metadata.timestamp = self.timestamp
        if self.crc_validity is not None:
            packet.metadata.is_crc_valid = self.crc_validity
        if self.relative_timestamp is not None:
            packet.metadata.relative_timestamp = self.relative_timestamp
        packet.metadata.decrypted = self.decrypted
        return packet

    @staticmethod
    def from_packet(packet):
        """Create message from Scapy packet
        """

        if BTLE in packet:
            # Extract PDU
            if BTLE_DATA in packet:
                pdu = raw(packet[BTLE_DATA:])
            elif BTLE_CTRL in packet:
                pdu = raw(packet[BTLE_CTRL:])
            elif BTLE_ADV in packet:
                pdu = raw(packet[BTLE_ADV:])
            else:
                return None

            return BleRawPduReceived(
                pdu=pdu,
                access_address=BTLE(raw(packet)).access_addr,
                crc=BTLE(raw(packet)).crc,
                direction=packet.metadata.direction,
                conn_handle=packet.metadata.connection_handle,
                channel=packet.metadata.channel,
                rssi=packet.metadata.rssi,
                timestamp=packet.metadata.timestamp,
                crc_validity=packet.metadata.is_crc_valid,
                relative_timestamp=packet.metadata.relative_timestamp,
                decrypted=packet.metadata.decrypted,
                processed=packet.metadata.processed
            )

        return None

@pb_bind(BleDomain, "injected", 1)
class Injected(PbMessageWrapper):
    """BLE PDU injected notification message class
    """
    success = PbFieldBool("ble.injected.success")
    access_address = PbFieldInt("ble.injected.access_address")
    injection_attempts = PbFieldInt("ble.injected.injection_attempts")



"""Bluetooth Low Energy Battery Service Profile
"""
from struct import pack, unpack
from whad.ble.profile.attribute import UUID
from whad.ble.profile.characteristic import Characteristic, Properties
from whad.ble.profile.service import StandardService

class BatteryService(StandardService):
    """Battery Service Profile

    This service implements the BLE Battery Service as described in the spec.
    """
    _uuid = UUID(0x180f)

    level = Characteristic(
        uuid = UUID(0x2A19),
        properties = Properties.READ | Properties.NOTIFY,
        required = True,
        value=pack('B', 100),
    )

    def __init__(self, handle: int = 0, end_handle: int = 0):
        super().__init__(
            handle,
            end_handle
        )

    @property
    def percentage(self) -> int:
        """Battery level as percentage."""
        return unpack('B', self.level.value)[0]

    @percentage.setter
    def percentage(self, level: int):
        """Set battery level."""
        if 0 <= level <=100:
            # level property is dynamically created by the GATT
            # profile where this class is used.
            #
            # pylint: disable-next=E1101
            self.level.value = pack('B', level)




from whad.rf4ce.stack.apl.profile import APLProfile
from whad.scapy.layers.rf4ce import RC_COMMAND_CODES, RF4CE_Vendor_MSO_Hdr, \
    RF4CE_Vendor_MSO_Audio, RF4CE_Vendor_MSO_Audio_Data_Notify, \
    RF4CE_Vendor_MSO_Audio_Stop_Request, RF4CE_Vendor_MSO_Audio_Stop_Response, \
    RF4CE_Vendor_MSO_Audio_Start_Request, RF4CE_Vendor_MSO_Audio_Start_Response, \
    RF4CE_Vendor_MSO_User_Control_Pressed, RF4CE_Vendor_MSO_User_Control_Released, \
    RF4CE_Vendor_MSO_User_Control_Repeated, RF4CE_Vendor_MSO_Get_Attribute_Request, \
    RF4CE_Vendor_MSO_Get_Attribute_Response, RF4CE_Vendor_MSO_Check_Validation_Request, \
    RF4CE_Vendor_MSO_Check_Validation_Response

from whad.rf4ce.stack.apl.exceptions import APLTimeoutException
from whad.rf4ce.stack.apl.profiles.mso.database import InformationBase
from whad.rf4ce.stack.apl.profiles.mso.parsers import TXOptionsValue
from whad.rf4ce.utils.adpcm import ADPCM

from enum import IntEnum
from time import sleep, time

class MSOEvent(IntEnum):
    """
    Enum representing types of MSO events.
    """
    DISCOVERY_REQ       = 1
    PAIRING_REQ         = 2
    GET_ATTRIBUTE_REQ   = 3
    VALIDATION_REQ      = 4
    KEY_PRESSED         = 5
    KEY_RELEASED        = 6
    KEY_REPEATED        = 7
    AUDIO_START_REQ     = 8
    AUDIO_STOP_REQ      = 9
    AUDIO_DATA          = 10
    VALIDATION_RSP      = 11

class MSOProfile(APLProfile):
    """
    APL service implementing the MSO Profile.
    """

    def __init__(self):
        self.information_base = InformationBase()
        self.validation = False
        self.keystrokes = False
        self.audio = None
        super().__init__(name="mso", profile_id=0xC0)

    def init(self):
        """
        Initialize the NWK layer database according to MSO profile.
        """
        nwk_layer = self.manager.get_layer('nwk')
        nwk_layer.database.set("nwkDiscoveryLQIThreshold", 0) # no filtering at NWK layer
        nwk_layer.database.set("nwkDiscoveryRepetitionInterval", 0x00927c) # 600ms
        nwk_layer.database.set("nwkIndicateDiscoveryRequest", True)
        nwk_layer.database.set("nwkMaxDiscoveryRepetitions", 2)
        nwk_layer.database.set("nwkMaxReportedNodeDescriptors", 16)


    def send_audio(self, audio_filename="/tmp/trololo.wav"):
        """
        Transmit audio packets.
        """
        nwk_layer = self.manager.get_layer('nwk')

        for p in ADPCM.convert_input_file(audio_filename):
            pkt = RF4CE_Vendor_MSO_Hdr()/RF4CE_Vendor_MSO_Audio()/p

            nwk_layer.get_service('data').data(
                pkt,
                pairing_reference = self.pairing_reference,
                profile_id = self.profile_id,
                vendor_id = self.manager.get_layer('nwk').database.get("nwkVendorIdentifier"),
                tx_options = (
                    TXOptionsValue.pack(
                        broadcast_transmission = False,
                        extended_address_mode = False,
                        acknowledgement_mode = False,
                        security_enabled = True,
                        channel_agility_mode = False,
                        channel_normalization_mode = False,
                        vendor_specific = True
                    )[0]
                )
            )



    def send_key(self, keystroke):
        """
        Transmit a key to target.
        """
        nwk_layer = self.manager.get_layer('nwk')
        code = None
        for key, value in RC_COMMAND_CODES.items():
            if value == keystroke:
                code = key
                break
        pkt = RF4CE_Vendor_MSO_Hdr()/RF4CE_Vendor_MSO_User_Control_Pressed(
            code= code
        )

        nwk_layer.get_service('data').data(
            pkt,
            pairing_reference = self.pairing_reference,
            profile_id = self.profile_id,
            vendor_id = self.manager.get_layer('nwk').database.get("nwkVendorIdentifier"),
            tx_options = (
                TXOptionsValue.pack(
                    broadcast_transmission = False,
                    extended_address_mode = False,
                    acknowledgement_mode = False,
                    security_enabled = True,
                    channel_agility_mode = False,
                    channel_normalization_mode = False,
                    vendor_specific = True
                )[0]
            )
        )

    def process_discovery(self, discovery):
        """
        Process a discovery request.
        """
        nwk_layer = self.manager.get_layer('nwk')
        apl_layer = self.manager.get_layer('apl')
        (
            status,
            source_address,
            node_capability,
            vendor_identifier,
            vendor_string,
            application_capability,
            user_string,
            device_type_list,
            profile_identifier_list,
            requested_device_type,
            link_quality
        ) = discovery

        # we don't match specification of MSO here, spoof vendor identifier instead of checking it
        nwk_layer.database.set("nwkVendorIdentifier", vendor_identifier)

        # we don't match specification of MSO here, spoof requested device type instead of checking it
        if requested_device_type == 0xFF:
            apl_layer.database.set("aplDeviceType", 9) # default
        else:
            apl_layer.database.set("aplDeviceType", requested_device_type)

        # we only filter by profile ID
        if self.profile_id not in profile_identifier_list:
            return

        list_of_supported_profiles = [profile_id for profile_id in self.manager.profiles.keys()]

        nwk_layer.get_service('management').discovery_response(
            0,
            source_address,
            list_of_device_types=[apl_layer.database.get("aplDeviceType")],
            list_of_profiles=list_of_supported_profiles,
            link_quality=link_quality
        )

    def pairing(self, address, pan_id):
        """
        Establish pairing.
        """
        nwk_layer = self.manager.get_layer('nwk')
        pairing_reference = nwk_layer.get_service('management').pair_request(
            destination_ieee_address=address,
            destination_pan_id=pan_id
        )
        return pairing_reference


    def bind(self, address, pan_id, timeout=30, validation_code_callback = input):

        self.pairing_reference = self.pairing(address, pan_id)

        nwk_layer = self.manager.get_layer('nwk')

        pkt = RF4CE_Vendor_MSO_Hdr()/RF4CE_Vendor_MSO_Get_Attribute_Request(
            attribute_identifier = 0xDC,
            index = 0,
            value_length = 4
        )

        nwk_layer.get_service('data').data(
            pkt,
            pairing_reference = self.pairing_reference,
            profile_id = self.profile_id,
            vendor_id = self.manager.get_layer('nwk').database.get("nwkVendorIdentifier"),
            tx_options = (
                TXOptionsValue.pack(
                    broadcast_transmission = False,
                    extended_address_mode = False,
                    acknowledgement_mode = False,
                    security_enabled = True,
                    channel_agility_mode = False,
                    channel_normalization_mode = False,
                    vendor_specific = True
                )[0]
            )
        )

        self.validation = True

        while True:
            pkt = RF4CE_Vendor_MSO_Hdr()/RF4CE_Vendor_MSO_Check_Validation_Request(
                request_automatic_validation = "yes"
            )

            nwk_layer.get_service('data').data(
                pkt,
                pairing_reference = self.pairing_reference,
                profile_id = self.profile_id,
                vendor_id = self.manager.get_layer('nwk').database.get("nwkVendorIdentifier"),
                tx_options = (
                    TXOptionsValue.pack(
                        broadcast_transmission = False,
                        extended_address_mode = False,
                        acknowledgement_mode = False,
                        security_enabled = True,
                        channel_agility_mode = False,
                        channel_normalization_mode = False,
                        vendor_specific = True
                    )[0]
                )
            )
            start_time = time()
            while (time() - start_time) < 1:

                try:
                    event = self.wait_for_packet(lambda _ : True)
                    if event[0] == MSOEvent.VALIDATION_RSP:
                        if event[1].check_validation_status == 0:
                            return True
                        elif event[1].check_validation_status > 0xc0:
                            return False
                except APLTimeoutException:
                    pass

                if validation_code_callback is not None:
                    for i in str(validation_code_callback()):
                        self.send_key(i)

    def process_pairing(self, pairing):
        '''
        Process a pairing event.
        '''
        (
            status,
            source_pan_id,
            source_address,
            node_capability,
            vendor_identifier,
            vendor_string,
            application_capability,
            user_string,
            device_type_list,
            profile_identifier_list,
            key_exchange_transfer_count,
            pairing_reference
        ) = pairing

        mac_layer = self.manager.get_layer('mac')
        nwk_layer = self.manager.get_layer('nwk')
        apl_layer = self.manager.get_layer('apl')

        nwk_layer.database.set("nwkUserString", apl_layer.database.get("aplUserString"))
        list_of_supported_profiles = [profile_id for profile_id in self.manager.profiles.keys()]
        list_of_device_types = [apl_layer.database.get("aplDeviceType")]


        if nwk_layer.get_service('management').pair_response(
            source_address,
            accept=True,
            list_of_device_types=[apl_layer.database.get("aplDeviceType")],
            list_of_profiles=list_of_supported_profiles,
            pairing_reference=pairing_reference
        ):
            return pairing_reference
        else:
            return None

    def wait_for_binding(self, timeout=30):
        """
        Allow binding until timeout is reached.
        """
        start_time = time()

        while (time() - start_time) < timeout:
            try:
                event = self.wait_for_packet(lambda _ : True)
                if event[0] == MSOEvent.DISCOVERY_REQ:
                    self.process_discovery(event[1:])
                elif event[0] == MSOEvent.PAIRING_REQ:
                    reference = self.process_pairing(event[1:])
                    if reference is None:
                        return None
                    else:
                        self.pairing_reference = reference
                        return self.manager.get_layer('nwk').database.get("nwkPairingTable")[reference]
            except APLTimeoutException:
                pass
        return None

    def wait_for_validation_code(self, code, timeout=30):
        self.keystrokes = True
        self.validation = True

        if isinstance(code, int):
            validation_code = str(code)
        elif isinstance(code, str):
            validation_code = code
        else:
            return False

        start_time = time()

        while (time() - start_time) < timeout:
            try:
                event = self.wait_for_packet(lambda _ : True)
                if event[0] == MSOEvent.KEY_PRESSED:
                    if validation_code.startswith(event[1][0]):
                        validation_code = validation_code[1:]
                        if len(validation_code) == 0:
                            self.keystrokes = False
                            return True
                    else:
                        self.keystrokes = False
                        return False

                elif event[0] == MSOEvent.VALIDATION_REQ:
                    pkt = RF4CE_Vendor_MSO_Hdr()/RF4CE_Vendor_MSO_Check_Validation_Response(
                        check_validation_status = 0xC0 # pending
                    )

                    nwk_layer = self.manager.get_layer('nwk')

                    nwk_layer.get_service('data').data(
                        pkt,
                        pairing_reference = self.pairing_reference,
                        profile_id = self.profile_id,
                        vendor_id = self.manager.get_layer('nwk').database.get("nwkVendorIdentifier"),
                        tx_options = (
                            TXOptionsValue.pack(
                                broadcast_transmission = False,
                                extended_address_mode = True,
                                acknowledgement_mode = False,
                                security_enabled = True,
                                channel_agility_mode = False,
                                channel_normalization_mode = False,
                                vendor_specific = True
                            )[0]
                        )
                    )
            except APLTimeoutException:
                pass

        self.keystrokes = False
        return False

    def accept_validation(self):
        """
        Accept validation phase.
        """
        self.validation = True

        while True:
            try:
                event = self.wait_for_packet(lambda _ : True)
                if event[0] == MSOEvent.VALIDATION_REQ:
                    pkt = RF4CE_Vendor_MSO_Hdr()/RF4CE_Vendor_MSO_Check_Validation_Response(
                        check_validation_status = 0x00
                    )

                    nwk_layer = self.manager.get_layer('nwk')

                    for _ in range(3):
                        nwk_layer.get_service('data').data(
                            pkt,
                            pairing_reference = self.pairing_reference,
                            profile_id = self.profile_id,
                            vendor_id = self.manager.get_layer('nwk').database.get("nwkVendorIdentifier"),
                            tx_options = (
                                TXOptionsValue.pack(
                                    broadcast_transmission = False,
                                    extended_address_mode = True,
                                    acknowledgement_mode = False,
                                    security_enabled = True,
                                    channel_agility_mode = False,
                                    channel_normalization_mode = False,
                                    vendor_specific = True
                                )[0]
                            )
                        )
                    return True

            except APLTimeoutException:
                pass
        return False

    def deny_validation(self):
        """
        Deny validation phase.
        """
        self.validation = True

        while True:
            try:
                event = self.wait_for_packet(lambda _ : True)
                if event[0] == MSOEvent.VALIDATION_REQ:
                    pkt = RF4CE_Vendor_MSO_Hdr()/RF4CE_Vendor_MSO_Check_Validation_Response(
                        check_validation_status = 0xC3
                    )

                    nwk_layer = self.manager.get_layer('nwk')

                    for _ in range(3):
                        nwk_layer.get_service('data').data(
                            pkt,
                            pairing_reference = self.pairing_reference,
                            profile_id = self.profile_id,
                            vendor_id = self.manager.get_layer('nwk').database.get("nwkVendorIdentifier"),
                            tx_options = (
                                TXOptionsValue.pack(
                                    broadcast_transmission = False,
                                    extended_address_mode = True,
                                    acknowledgement_mode = False,
                                    security_enabled = True,
                                    channel_agility_mode = False,
                                    channel_normalization_mode = False,
                                    vendor_specific = True
                                )[0]
                            )
                        )
                    return True

            except APLTimeoutException:
                pass
        return False

    def key_stream(self):
        self.keystrokes = True
        while True:
            try:
                event = self.wait_for_packet(lambda _ : True)
                if event[0] == MSOEvent.KEY_PRESSED:
                    yield event[1]
            except APLTimeoutException:
                pass

    def audio_stream(self):
        while True:
            try:
                event = self.wait_for_packet(lambda _ : True)
                if event[0] == MSOEvent.AUDIO_DATA:
                    yield event[2]
            except APLTimeoutException:
                pass

    def save_audio(self, filename):
        self.audio = ADPCM(output_filename = filename)

    def live_audio(self):
        self.audio = ADPCM(live_play = True)

    def on_audio_payload(self, payload):
        if self.audio is not None:
            self.audio.process_packet(payload)

        if RF4CE_Vendor_MSO_Audio_Start_Request in payload:
            self.add_packet_to_queue(
                (
                    MSOEvent.AUDIO_START_REQ,
                    payload.sample_rate,
                    payload.resolution_bits,
                    payload.mic_channel_number,
                    payload.codec_type,
                    payload.packet_size,
                    payload.interval,
                    payload.channel_number,
                    payload.duration,
                )
            )
            pkt = RF4CE_Vendor_MSO_Hdr()/RF4CE_Vendor_MSO_Audio()/RF4CE_Vendor_MSO_Audio_Start_Response()

            nwk_layer = self.manager.get_layer('nwk')
            nwk_layer.get_service('data').data(
                pkt,
                pairing_reference = self.pairing_reference,
                profile_id = self.profile_id,
                vendor_id = self.manager.get_layer('nwk').database.get("nwkVendorIdentifier"),
                tx_options = (
                    TXOptionsValue.pack(
                        broadcast_transmission = False,
                        extended_address_mode = True,
                        acknowledgement_mode = False,
                        security_enabled = True,
                        channel_agility_mode = False,
                        channel_normalization_mode = False,
                        vendor_specific = True
                    )[0]
                )
            )

        elif RF4CE_Vendor_MSO_Audio_Stop_Request in payload:
            self.add_packet_to_queue(
                (
                    MSOEvent.AUDIO_STOP_REQ,
                )
            )
            pkt = RF4CE_Vendor_MSO_Hdr()/RF4CE_Vendor_MSO_Audio()/RF4CE_Vendor_MSO_Audio_Stop_Response()

            nwk_layer = self.manager.get_layer('nwk')
            nwk_layer.get_service('data').data(
                pkt,
                pairing_reference = self.pairing_reference,
                profile_id = self.profile_id,
                vendor_id = self.manager.get_layer('nwk').database.get("nwkVendorIdentifier"),
                tx_options = (
                    TXOptionsValue.pack(
                        broadcast_transmission = False,
                        extended_address_mode = True,
                        acknowledgement_mode = False,
                        security_enabled = True,
                        channel_agility_mode = False,
                        channel_normalization_mode = False,
                        vendor_specific = True
                    )[0]
                )
            )
        elif RF4CE_Vendor_MSO_Audio_Data_Notify in payload:
            self.add_packet_to_queue(
                (
                    MSOEvent.AUDIO_DATA,
                    payload.header,
                    payload.samples
                )
            )

    def on_user_control_pressed_payload(self, payload):
        """
        Callback processing user control pressed payload.
        """
        if payload.code in RC_COMMAND_CODES:
            print("[i] MSO - control pressed - ", RC_COMMAND_CODES[payload.code])
        else:
            print("[i] MSO - control pressed - unknown")

        if self.keystrokes:
            self.add_packet_to_queue(
                (
                    MSOEvent.KEY_PRESSED,
                    RC_COMMAND_CODES[payload.code]
                )
            )

    def on_user_control_repeated_payload(self, payload):
        """
        Callback processing user control repeated payload.
        """
        print("[i] MSO - control repeated")

        if self.keystrokes:
            self.add_packet_to_queue(
                (
                    MSOEvent.KEY_REPEATED
                )
            )

    def on_user_control_released_payload(self, payload):
        """
        Callback processing user control released payload.
        """
        print("[i] MSO - control released")

        if self.keystrokes:
            self.add_packet_to_queue(
                (
                    MSOEvent.KEY_RELEASED
                )
            )

    def on_check_validation_request_payload(self, payload):
        """
        Callback processing check validation request payload.
        """
        if self.validation:
            self.add_packet_to_queue(
                (
                    MSOEvent.VALIDATION_REQ,
                    payload
                )
            )
        return


    def on_check_validation_response_payload(self, payload):
        """
        Callback processing check validation response payload.
        """
        if self.validation:
            self.add_packet_to_queue(
                (
                    MSOEvent.VALIDATION_RSP,
                    payload
                )
            )
        return



    def on_get_attribute_response_payload(self, payload):
        """
        Callback processing Get Attribute Response payload.
        """
        print("[i] Get attribute response:", payload.value.hex())

    def on_get_attribute_request_payload(self, payload):
        """
        Callback processing Get Attribute Request payload.
        """
        attribute = self.information_base.get(lambda attribute : attribute.identifier == payload.attribute_identifier)
        if attribute is not None:
            value = attribute.value[payload.index:payload.index + payload.value_length]
            status = 0
        else:
            value = b""
            status = 1

        pkt = RF4CE_Vendor_MSO_Hdr()/RF4CE_Vendor_MSO_Get_Attribute_Response(
            attribute_identifier = payload.attribute_identifier,
            index = payload.index,
            status = status,
            value = value,
            value_length = len(value)
        )

        nwk_layer = self.manager.get_layer('nwk')
        nwk_layer.get_service('data').data(
            pkt,
            pairing_reference = self.pairing_reference,
            profile_id = self.profile_id,
            vendor_id = self.manager.get_layer('nwk').database.get("nwkVendorIdentifier"),
            tx_options = (
                TXOptionsValue.pack(
                    broadcast_transmission = False,
                    extended_address_mode = True,
                    acknowledgement_mode = False,
                    security_enabled = True,
                    channel_agility_mode = False,
                    channel_normalization_mode = False,
                    vendor_specific = True
                )[0]
            )
        )

    def on_data(self, npdu, pairing_reference, vendor_id, link_quality, rx_flags):
        """
        Callback processing incoming data for the profile.
        """

        if RF4CE_Vendor_MSO_User_Control_Pressed in npdu:
            self.on_user_control_pressed_payload(npdu[RF4CE_Vendor_MSO_User_Control_Pressed])

        elif RF4CE_Vendor_MSO_User_Control_Released in npdu:
            self.on_user_control_released_payload(npdu[RF4CE_Vendor_MSO_User_Control_Released])

        elif RF4CE_Vendor_MSO_User_Control_Repeated in npdu:
            self.on_user_control_repeated_payload(npdu[RF4CE_Vendor_MSO_User_Control_Repeated])

        elif RF4CE_Vendor_MSO_Get_Attribute_Request in npdu:
            self.on_get_attribute_request_payload(npdu[RF4CE_Vendor_MSO_Get_Attribute_Request])

        elif RF4CE_Vendor_MSO_Get_Attribute_Response in npdu:
            self.on_get_attribute_response_payload(npdu[RF4CE_Vendor_MSO_Get_Attribute_Response])

        elif RF4CE_Vendor_MSO_Check_Validation_Request in npdu:
            self.on_check_validation_request_payload(npdu[RF4CE_Vendor_MSO_Check_Validation_Request])
        elif RF4CE_Vendor_MSO_Check_Validation_Response in npdu:
            self.on_check_validation_response_payload(npdu[RF4CE_Vendor_MSO_Check_Validation_Response])
        elif RF4CE_Vendor_MSO_Audio in npdu:
            self.on_audio_payload(npdu[RF4CE_Vendor_MSO_Audio])

    def on_discovery(self, status, source_address, node_capability, vendor_identifier, vendor_string, application_capability, user_string, device_type_list, profile_identifier_list, requested_device_type, link_quality):
        """
        Callback processing incoming discovery request for the profile.
        """
        self.add_packet_to_queue(
            (
                MSOEvent.DISCOVERY_REQ,
                status,
                source_address,
                node_capability,
                vendor_identifier,
                vendor_string,
                application_capability,
                user_string,
                device_type_list,
                profile_identifier_list,
                requested_device_type,
                link_quality
            )
        )




    def on_pair(self, status, source_pan_id, source_address, node_capability, vendor_identifier, vendor_string, application_capability, user_string, device_type_list, profile_identifier_list, key_exchange_transfer_count, pairing_reference):
        """
        Callback processing incoming pairing request for the profile.
        """
        print("[i] MSO - pairing request")
        self.add_packet_to_queue(
            (
                MSOEvent.PAIRING_REQ,
                status,
                source_pan_id,
                source_address,
                node_capability,
                vendor_identifier,
                vendor_string,
                application_capability,
                user_string,
                device_type_list,
                profile_identifier_list,
                key_exchange_transfer_count,
                pairing_reference
            )
        )



'''
This is the base class representing a event occuring during sniffing.
'''

class EventsManager:
    def __init__(self):
        self.__listeners = []

    def add_event_listener(self, listener):
        self.__listeners.append(listener)

    def trigger_event(self, event):
        for listener in self.__listeners:
            listener(event)

class SniffingEvent:
    def __init__(self, name):
        self.name = name

    @property
    def message(self):
        return None



"""
This module provides the :class:`whad.unifying.connector.sniffer.Sniffer` class
that allows Logitech Unifying packets sniffing.



"""
import logging

from time import time
from typing import Generator, List
from scapy.packet import Packet

from whad.device import WhadDevice
from whad.exceptions import WhadDeviceDisconnected
from whad.unifying.connector import Unifying
from whad.unifying.sniffing import SnifferConfiguration, KeyExtractedEvent
from whad.unifying.crypto import LogitechUnifyingDecryptor, LogitechUnifyingKeyDerivation
from whad.exceptions import UnsupportedCapability
from whad.helpers import message_filter
from whad.hub.unifying import RawPduReceived, PduReceived
from whad.common.sniffing import EventsManager
from whad.scapy.layers.unifying import Logitech_Unifying_Hdr, Logitech_Encrypted_Keystroke_Payload
from whad.hub.message import AbstractPacket

logger = logging.getLogger(__name__)

class Sniffer(Unifying, EventsManager):
    """
    Logitech Unifying Sniffer interface for compatible WHAD device.
    """

    def __init__(self, device: WhadDevice):
        """Sniffer initialization.

        :param device: WHAD device to use
        :type device: :class:`whad.device.WhadDevice`
        """
        Unifying.__init__(self, device)
        EventsManager.__init__(self)

        self.__configuration = SnifferConfiguration()
        self.__decryptor = LogitechUnifyingDecryptor()
        self.__key_derivation = LogitechUnifyingKeyDerivation()

        self.__addresses = []
        # Check if device can perform sniffing
        if not self.can_sniff():
            raise UnsupportedCapability("Sniff")


    def _enable_sniffing(self):
        """Enable sniffing.

        This method takes the current sniffing configuration and configure the
        hardware accordingly.
        """
        for key in self.__configuration.keys:
            self.__decryptor.add_key(key)
        if self.__configuration.scanning:
            channel = None
        else:
            channel = self.__configuration.channel

        ack = self.__configuration.acknowledgements
        address = self.__configuration.address

        if self.__configuration.pairing:
            self.sniff_pairing()
        else:
            super().start_sniff(channel=channel, show_acknowledgements=ack, address=address)

    @property
    def configuration(self):
        """Current sniffing configuration accessor.
        """
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration) -> SnifferConfiguration:
        """Sniffing configuration setter.

        :return: Sniffing configuration object
        :rtype: :class:`whad.unifying.sniffing.SnifferConfiguration`
        """
        self.stop()
        self.__configuration = new_configuration
        self._enable_sniffing()

    @property
    def channel(self) -> int:
        """Retrieve current channel.

        :return: current channel
        :rtype: int
        """
        return self.__configuration.channel

    @channel.setter
    def channel(self, channel: int = 11):
        """Sniffing channel setter.

        :param channel: Channel to use
        :type channel: int
        """
        self.stop()
        self.__configuration.channel = channel
        self._enable_sniffing()

    def add_key(self, key: bytes):
        """Add encryption key to the curret sniffing configuration.

        :param key: Key to add
        :type key: bytes
        """
        self.stop()
        self.__configuration.keys.append(key)
        self._enable_sniffing()

    def clear_keys(self):
        """Remove all registered keys from current sniffing configuration.
        """
        self.stop()
        self.__configuration.keys = []
        self._enable_sniffing()

    @property
    def decrypt(self):
        """Current sniffing decryption flag accessor.
        """
        return self.__configuration.decrypt

    @decrypt.setter
    def decrypt(self, decrypt: bool):
        """Enable or disable decryption in current sniffing configuration.
        """
        self.stop()
        self.__configuration.decrypt = decrypt
        self._enable_sniffing()

    @property
    def address(self):
        """Target address accessor.
        """
        return self.__configuration.address

    @address.setter
    def address(self, address: str):
        """Target address setter.

        :param address: target address
        :type address: str
        """
        self.stop()
        self.__configuration.address = address
        self._enable_sniffing()

    @property
    def scanning(self):
        """Current sniffing configuration scanning flag accessor.
        """
        return self.__configuration.scanning

    @scanning.setter
    def scanning(self, scanning: bool):
        """Enable or disable scanning.

        :param scanning: Enable scanning if set to True, disable it if False
        :type scanning: bool
        """
        self.stop()
        self.__configuration.scanning = scanning
        self._enable_sniffing()

    def available_actions(self, filter=None) -> List:
        """Identify available actions.
        """
        actions = []
        return [action for action in actions if filter is None or isinstance(action, filter)]

    def process_packet(self, packet):
        """Implement the packet decryption if needed.
        """

        if self.__configuration.pairing:
            self.__key_derivation.process_packet(packet)
            if self.__key_derivation.key is not None:
                logger.info("[i] New key extracted: ", self.__key_derivation.key.hex())
                self.trigger_event(KeyExtractedEvent(self.__key_derivation.key))
                self.add_key(self.__key_derivation.key)
                self.__key_derivation.reset()

        if Logitech_Encrypted_Keystroke_Payload in packet and self.__configuration.decrypt:
            decrypted, success = self.__decryptor.attempt_to_decrypt(packet)
            if success:
                packet.decrypted = decrypted
                # Replace packet if decrypted
                decrypted_packet = packet.copy()
                decrypted_packet.metadata = packet.metadata
                decrypted_packet[Logitech_Unifying_Hdr].remove_payload()
                decrypted_packet.payload = decrypted
                packet = decrypted_packet
                # packet.checksum = None
                packet.metadata = decrypted_packet.metadata
                packet.metadata.decrypted = True
                return packet

        return packet


    def sniff(self, timeout: float = None) -> Generator[Packet, None, None]:
        """Sniff Logitech Unifying packets

        :param timeout: Number of seconds after which sniffing stops, uninterrupted if set to None
        :type timeout: float
        """
        # Determine message type
        if self.support_raw_pdu():
            message_type = RawPduReceived
        else:
            message_type = PduReceived

        # Make sure we are s
        self._enable_sniffing()

        # Sniff packets
        start = time()

        for message in super().sniff(messages=(message_type), timeout=timeout):
            if message is not None and issubclass(message, AbstractPacket):
                packet = message.to_packet()
                if packet is not None:
                    packet = self.process_packet(packet)
                    self.monitor_packet_rx(packet)
                    yield packet



"""RF4CE Crypto manager
"""
from struct import pack
from scapy.config import conf

from Cryptodome.Cipher import AES
from Cryptodome.Random import get_random_bytes

from scapy.layers.dot15d4 import Dot15d4, Dot15d4FCS

from whad.common.analyzer import TrafficAnalyzer
from whad.scapy.layers.rf4ce import RF4CE_Hdr, \
    RF4CE_Cmd_Pair_Request, RF4CE_Cmd_Key_Seed

from .exceptions import MissingRF4CEHeader, \
    MissingCryptographicMaterial, MissingRF4CESecurityFlag

conf.dot15d4_protocol = "rf4ce"

def generate_random_value(bits):
    """Generate a random value of provided bit size.

    :param int bits: Number of bits (8-bit aligned) to generate.
    :return bytes: Random bytes
    """
    return get_random_bytes(int(bits/8))


class RF4CECryptoManager:
    def __init__(self, key):
        self.key = key
        self.nonce = None
        self.auth = None

    def generateNonce(self, packet, source=None):
        # Check source validity
        if source is None:
            if packet.fcf_srcaddrmode == 3: # Long address
                source = pack("<Q", packet.src_addr)

        if source is None:
            return None

        frame_counter = pack("<I", packet.frame_counter)
        security_level = b"\x05" # always 5 in RF4CE
        # build and return the nonce
        return source + frame_counter + security_level

    def generateAuth(self, packet, destination=None):
        # Check destination validity
        if destination is None:
            if packet.fcf_destaddrmode == 3: # Long address
                destination = pack("<Q", packet.dest_addr)

        if destination is None:
            return None

        # Extract the header (force a rebuild)
        packet.reserved = 1
        packet = packet[RF4CE_Hdr:]
        header = bytes(packet)[:5]

        # Build and return the auth
        return header + destination

    def extractCiphertextPayload(self, packet):
        # If it is a vendor or a data packet, we need to take
        # into account supplementary fields
        packet.reserved = 1
        (packet.do_build())
        start_of_payload = 2 if packet.frame_type in (1,3) else 1
        ciphertext = bytes(packet[RF4CE_Hdr:][start_of_payload:])
        #print(ciphertext.hex())
        return ciphertext, pack("<I", packet.mic)

    def extractPlaintextPayload(self, packet):
        packet.reserved = 1
        (packet.do_build())
        start_of_payload = 2 if packet.frame_type in (1,3) else 1
        plaintext = bytes(packet[RF4CE_Hdr:][start_of_payload:])
        return plaintext

    def decrypt(self, packet, source=None, destination=None, rf4ce_only=False):
        fcs_present = False
        # convert source and destination address if provided
        if isinstance(source, str) and ":" in source:
            source = bytes.fromhex(source.replace(":", ""))[::-1]

        if isinstance(destination, str) and ":" in destination:
            destination = bytes.fromhex(destination.replace(":", ""))[::-1]

        if rf4ce_only:
            if RF4CE_Hdr in packet:
                packet.reserved = 1
                packet = RF4CE_Hdr(packet.do_build())

            if isinstance(packet, bytes):
                packet = RF4CE_Hdr(packet)
        else:
            # convert into scapy packet if bytes only
            if isinstance(packet, bytes):
                packet = Dot15d4(packet)

            # don't process FCS if present
            if Dot15d4FCS in packet:
                # force a rebuild just in case
                packet.reserved = 1
                packet = packet.do_build()[:-2]
                packet = Dot15d4(packet)

                fcs_present = True

        if RF4CE_Hdr not in packet:
            raise MissingRF4CEHeader()

        # Check if packet has security enabled
        if (
            hasattr(packet, "security_enabled")# and
            #packet.security_enabled == 1
        ):
            self.nonce = self.generateNonce(packet, source)
            if self.nonce is None:
                # Missing source address, nonce cannot be generated
                return (packet, False)

            self.auth = self.generateAuth(packet, destination)
            if self.auth is None:
                # Missing destination address, auth cannot be generated
                return (packet, False)

            ciphertext, mic = self.extractCiphertextPayload(packet)
            # Perform the decryption and integrity check
            cipher = AES.new(self.key, AES.MODE_CCM, nonce=self.nonce, mac_len=4)
            cipher.update(self.auth)
            plaintext = cipher.decrypt(ciphertext)
            try:
                cipher.verify(mic)
                # Rebuild the decrypted packet
                header = bytes(packet)[:-4-len(ciphertext)]
                if rf4ce_only:
                    packet = RF4CE_Hdr(header + plaintext + mic)
                else:
                    packet = Dot15d4(header + plaintext + mic)
                    if fcs_present:
                        # rebuild Dot15d4FCS packet
                        packet = Dot15d4FCS(bytes(packet) + Dot15d4FCS.compute_fcs(None, bytes(packet)))

                return (packet, True)

            except ValueError:
                return (packet, False) # integrity check

        else:
            # Security is not enabled, raise an exception
            raise MissingRF4CESecurityFlag()


    def encrypt(self, packet, source=None, destination=None, rf4ce_only=False):
        fcs_present = False
        # convert source and destination address if provided
        if isinstance(source, str) and ":" in source:
            source = bytes.fromhex(source.replace(":", ""))[::-1]

        if isinstance(destination, str) and ":" in destination:
            destination = bytes.fromhex(destination.replace(":", ""))[::-1]

        if rf4ce_only:
            if RF4CE_Hdr in packet:
                packet.reserved = 1
                packet = RF4CE_Hdr(packet.do_build())

            if isinstance(packet, bytes):
                packet = RF4CE_Hdr(packet)
        else:
            # convert into scapy packet if bytes only
            if isinstance(packet, bytes):
                packet = Dot15d4(packet)

            # don't process FCS if present
            if Dot15d4FCS in packet:
                fcs_present = True
                # force a rebuild just in case
                packet.reserved = 1
                packet = bytes(packet)[:-2]
                packet = Dot15d4(packet)

            if RF4CE_Hdr not in packet:
                raise MissingRF4CEHeader()

        # Check if packet has security enabled
        if (
            hasattr(packet, "security_enabled")# and
            #packet.security_enabled == 1
        ):
            self.nonce = self.generateNonce(packet, source)
            if self.nonce is None:
                # Missing source address, nonce cannot be generated
                return (packet, False)

            self.auth = self.generateAuth(packet, destination)
            if self.auth is None:
                # Missing destination address, auth cannot be generated
                return (packet, False)

            # Perform the decryption and integrity check
            cipher = AES.new(self.key, AES.MODE_CCM, nonce=self.nonce, mac_len=4)
            cipher.update(self.auth)

            plaintext = self.extractPlaintextPayload(packet)
            ciphertext = cipher.encrypt(plaintext)
            mic = cipher.digest()

            if packet.mic is None:
                cropping_length = len(plaintext)
            else:
                cropping_length = len(plaintext) + 4

            packet.reserved = 1
            header = bytes(packet)[:-cropping_length]

            if rf4ce_only:
                packet = RF4CE_Hdr(header + ciphertext + mic)
            else:
                packet = Dot15d4(header + ciphertext +  mic)
                if fcs_present:
                    # rebuild Dot15d4FCS packet
                    packet = Dot15d4FCS(bytes(packet) + Dot15d4FCS.compute_fcs(None, bytes(packet)))

            # Set the security flag and force a rebuild
            packet.security_enabled = 1

            return packet

        else:
            # Security is not enabled, raise an exception
            raise MissingRF4CESecurityFlag()

def xor(a, b):
    """
    Helper function to perform a xor operation between two bytes array.
    """
    return bytes(
        [
            a[i] ^ b[i] for i in range(min(len(a), len(b)))
        ]
    )

class RF4CEKeyDerivation(TrafficAnalyzer):
    def __init__(self, seeds_number=None, seeds=[]):
        self.reset()
        self.seeds_number = seeds_number
        self.seeds = seeds

    def process_packet(self, packet):
        if RF4CE_Cmd_Pair_Request in packet:
            self.trigger()
            self.mark_packet(packet)
            self.seeds_number = packet.key_exchange_transfer_count
        elif RF4CE_Cmd_Key_Seed in packet:
            self.trigger()
            self.mark_packet(packet)
            self.seeds.append(packet.seed_data)

        if (
                self.seeds_number is not None and
                len(self.seeds) == (1 + self.seeds_number)
        ):
            self.complete()

    @property
    def output(self):
        return {"key":self.key}

    @property
    def key(self):
        if (
                self.seeds_number is not None and
                len(self.seeds) == (1 + self.seeds_number)
        ):
            return self._generate_key()
        else:
            return None

    def reset(self):
        super().reset()
        self.seeds_number = None
        self.seeds = []

    def _generate_key(self):
        phase1 = b"\x00" * 80
        for current_seed in self.seeds:
            phase1 = xor(phase1, current_seed)

        phase2 = []
        for i in range(5):
            start = i*16
            end = (i+1)*16
            s = phase1[start:end]
            phase2.append(s)

        key = b"\x00"*16
        for current_seed in phase2:
            key = xor(key, current_seed)

        return key

class RF4CEDecryptor:
    def __init__(self, *keys):
        self.keys = list(keys)
        self.addresses = []

    def add_address(self, *addresses):
        for address in addresses:
            if isinstance(address, int):
                address = ":".join(["{:02x}".format(i) for i in pack('>Q', address)])
            if address not in self.addresses:
                self.addresses.append(address)

    def add_key(self, key):
        if isinstance(key, str):
            if len(key) == 16:
                key = key.encode('ascii')
            else:
                try:
                    key = bytes.fromhex(key.replace(":",""))
                except ValueError:
                    return False

        if not isinstance(key, bytes) or len(key) != 16:
            return False

        if key not in self.keys:
            self.keys.append(key)
            return True
        return False

    def addresses_combination(self):
        tmp_list = [
            ((a, b), (b, a))
            for idx, a in enumerate(self.addresses)
            for b in self.addresses[idx + 1:]
        ]

        address_list = []

        for entry in tmp_list:
            first_pair, second_pair = entry
            address_list += [first_pair, second_pair]

        return address_list

    def attempt_to_decrypt(self, packet):
        if (
            len(self.keys) == 0 or
            (
                len(self.addresses) == 0 and
                (
                    packet.fcf_srcaddrmode != 3 or
                    packet.fcf_destaddrmode != 3
                )
            )
        ):
            raise MissingCryptographicMaterial()

        if RF4CE_Hdr not in packet:
            return (None, False)

        if packet.security_enabled == 0:
            return (None, False)

        for key in self.keys:
            manager = RF4CECryptoManager(key)
            candidate_addresses = self.addresses_combination()

            for (source_address, destination_address) in candidate_addresses:
                try:
                    decrypted_packet, success = manager.decrypt(
                        packet,
                        source_address,
                        destination_address
                    )

                    if success:
                        return (decrypted_packet, True)

                except:
                    pass

        return (None, False)



"""
WHAD Unit Tests Procedures
==========================

This module provides a basic implementation of a test
procedure for mocks through the `Procedure` class. This
class is a simple state-machine that must be extended
to define any testing procedure including a single or
multiple steps.

A procedure has an initial state and three possible terminal
states:
- a state indicating that the procedure has successfully completed
  and returned a result (*DONE* state) or
- a state indicating that an error occurred during the procedure
  (*ERROR* state)
- a state indicating the procedure has been canceled (*CANCELED* state)

"""
import logging

from typing import Optional, Any, Callable, List, Type
from threading import Event

from scapy.packet import Packet

# Default module logger
logger = logging.getLogger(__name__)

def proc_state(state: int):
    """Decorator to specify a method as an handler for a specific state.

    :param state: State associated with the decorated method.
    """
    def _wrapper(f):
        setattr(f, "__state_handler", state)
        return f
    return _wrapper


class UnexpectedProcError(Exception):
    """Unexpected error occurred while executing a test procedure."""

class ProcTimeoutError(Exception):
    """Procedure has not completed in due time."""

class ProcedureState:
    """Default procedure states.

    `INITIAL` state is the default state used in procedures when
    they are instantiated. The `DONE` state corresponds to a procedure
    that has been successfully completed, while the `ERROR` state is used
    to indicate that an error has occurred.

    The `USER` state shall be used to implement any user-defined states.
    """
    UNDEFINED = -1
    INITIAL = 0
    DONE = 1
    ERROR = 2
    CANCELED = 3
    USER = 4



class ProcedureMetaclass(type):
    """Procedure metaclass used to automatically register handlers
    associated with a procedure class.

    This metaclass is used by the `Procedure` class to automatically
    register any method decorated with `proc_state` as an handler
    associated with a specific state.

    A dictionary of registered handlers is injected in the class
    based on this metaclass as `STATE_HANDLERS`, that is then used
    in the `Procedure` class and its derived classes to call these
    the correct handler each time the procedure enters a specific
    state.
    """

    def __new__(cls, name, bases, dct):
        """Look for state handlers and update class's state handlers dictionary."""
        # Iterate over dct to register decorated handlers
        handlers = {}
        for name,obj in dct.items():
            if hasattr(obj, "__state_handler"):
                handlers[getattr(obj, "__state_handler")] = obj

        # Inject handlers dictionnary if not defined
        dct["STATE_HANDLERS"] = handlers

        # Create instance
        return super().__new__(cls, name, bases, dct)

class Procedure(metaclass=ProcedureMetaclass):
    """WHAD generic procedure for mocks.

    This class defines a generic test procedure used in mocks. It is a basic
    state machine able to process incoming and outgoing packets independently
    of a specific protocol. It is designed to be used in mocks to interact with
    any WHAD device in order to implement speciic unit tests.

    Handlers associated with a given state can be declared in this class by using
    the `proc_state` decorator. Such handlers will be called each time the procedure
    enters their associated states.

    Procedure is considered completed when its state is one of the terminal states:
    `ProcedureState.DONE`, `ProcedureState.ERROR` or `ProcedureState.CANCELED`.
    Waiting for procedure completion can be achieved by calling its `wait()` method.

    States associated with the procedure class are defined in a pseudo-enum class
    deriving from `ProcedureState` and set in the procedure class's `StatesEnum`
    property.
    """
    # Handlers for various states
    STATE_HANDLERS = {}

    # Terminal states
    TERMINAL_STATES = (
        ProcedureState.DONE,
        ProcedureState.ERROR,
        ProcedureState.CANCELED
    )

    def __init__(self):
        """Initialize this procedure."""
        # Set default state
        self.__state = ProcedureState.UNDEFINED
        self.__result = None
        self.__completed = Event()

    @classmethod
    def get_states_enum(cls) -> Type[ProcedureState]:
        return ProcedureState

    @property
    def states(self) -> Type[ProcedureState]:
        """Associated procedure state enum."""
        return self.get_states_enum()

    @classmethod
    def get_handler(cls, state: int) -> Optional[Callable[[int, int], None]]:
        """Retrieve the handler associated with a given state, if any."""
        if state in cls.STATE_HANDLERS:
            return cls.STATE_HANDLERS[state]
        return None

    def start(self):
        """Initiate procedure."""
        self.set_state(ProcedureState.INITIAL)

    def call_handler(self, prev:int, state: int) -> bool:
        """Call handler for given state."""
        handler = self.get_handler(state)
        if handler is not None:
            handler(self, prev, state)
            return True
        else:
            return False

    def set_state(self, state: int):
        """Set current state and triggers state handlers if any.

        :param state: Procedure state to set
        :type state: int
        """
        # Update state
        prev_state = self.__state
        self.__state = state

        if state not in Procedure.TERMINAL_STATES:
            # Call handler on state change
            self.call_handler(prev_state, state)
        else:
            # Mark procedure as terminated
            self.__completed.set()

    def get_state(self) -> int:
        """Get current procedure state."""
        return self.__state

    def is_state(self, state: int) -> bool:
        """Check current state."""
        return self.__state == state

    def set_result(self, result: Any):
        """Set result for procedure."""
        self.__result = result

    def get_result(self) -> Any:
        """Procedure result."""
        return self.__result

    def success(self) -> bool:
        """Determine if the procedure has been successful."""
        return self.__state == ProcedureState.DONE

    def error(self) -> bool:
        """Determine if an error occurred during the procedure."""
        return self.__state == ProcedureState.ERROR

    def cancel(self):
        """Cancel current procedure."""
        self.set_result(None)
        self.set_state(ProcedureState.CANCELED)

    def canceled(self) -> bool:
        """Determine if the procedure has been canceled."""
        return self.__state == ProcedureState.CANCELED

    def wait(self, timeout: Optional[float] = None) -> Optional[Any]:
        """Wait for this procedure to terminate.

        :param timeout: Maximum time allowed for the procedure to complete
        :type timeout: float, optional
        :return: Procedure result
        :rtype: object, optional
        """
        if self.__completed.wait(timeout=timeout):
            return self.__result
        else:
            raise ProcTimeoutError()

class StackProcedure(Procedure):
    """
    Test procedure implementation for device mocks.

    This procedure base class allows to define a list of packets to be sent
    when the procedure is initiated, and to process incoming packets through
    its `process` method. Procedure state is updated when `process()` is called,
    until it reaches a final state (*DONE*, *ERROR* or *CANCEL*).
    """

    def __init__(self, packets: Optional[List[Packet]] = None):
        """Initialize stack procedure.

        :param packets: List of initial packets for this procedure
        :type packets: list, optional
        """
        self.__initial_packets = packets or []
        super().__init__()

    def initiate(self) -> List[Packet]:
        """Start stack procedure and return initial packets.

        :return: List of initial packets for this procedure
        :rtype: list
        """
        self.start()
        return self.__initial_packets

    def process(self, packet: Packet) -> List[Packet]:
        """Incoming packet processing callback.

        This callback can be overriden to update the procedure state
        depending on any received packet.

        :param packet: Received packet
        :type packet: Packet
        :return: List of packets to send in response of the received packet, if any
        :rtype: list
        """
        logger.debug("processing packet %s", packet)
        return []




"""
ESB Scanning device database
============================

This module provides a database that keeps track of discovered devices,
:class:`whad.esb.scanning.CommunicatingDevicesDB`. Discovered devices information
are handled in :class:`whad.ble.scanning.CommunicatingDevice`.
"""
from typing import List

from whad.scapy.layers.esb import ESB_Payload_Hdr, ESB_Ack_Response

from .stack.llm.constants import ESBRole
from .esbaddr import ESBAddress


UNIFYING_COMMAND_BYTES = (0x51,0xC2,0x40,0x4F,0xD3,0xC1, 0xC3,0x5F,0x1F,0x0F,0x0E,0x10)

class CommunicatingDevice:
    """Store information about a device:

    * Received Signal Strength Indicator (RSSI)
    * Role (ptx or prx)
    * Applicative layer (if identified)
    * Set of used channels
    """

    def __init__(self, rssi, address, role, applicative_layer=None, channel=None):
        """Instantiate a CommunicatingDevice.

        :param  rssi:                   Received Signal Strength Indicator
        :type   rssi:                   float
        :param  address:                Device address
        :type   address:                str
        :param  role:                   Device role
        :type   role:                   :class:`whad.esb.ESBRole`
        :param  applicative_layer:       Scan response data
        :type   applicative_layer:      str
        :param  channel:               Last used channel
        :type   channel:               int
        """
        self.__address = ESBAddress(address)
        self.__role = role
        self.__applicative_layer = applicative_layer
        self.__rssi = rssi
        self.__channels = [channel]

    @property
    def address(self) -> str:
        """Device address.
        """
        return str(self.__address)

    @property
    def rssi(self) -> float:
        """Device RSSI.
        """
        return self.__rssi

    @property
    def role(self) -> ESBRole:
        """Device Role (Primary Transmitter (PTX) or Primary Receiver (PRX)).
        """
        return self.__role

    @property
    def channels(self) -> List[int]:
        """Channels used by the device.

        :return: list of channels used by the device in ascending order
        :rtype: list
        """
        channels = list(set(self.__channels))
        channels.sort()
        return channels

    @property
    def last_channel(self) -> int:
        """Last channel used by the device.
        """
        if len(self.__channels) == 0:
            return None
        return self.__channels[-1]

    def __repr__(self):
        """Show device information.
        """
        # Do we have an applicative protocol identified ?
        if self.__applicative_layer is None:
            app_layer = ""
        else:
            app_layer = f"{self.__applicative_layer}"

        # Display device role
        if self.__role == ESBRole.PRX:
            role = "[PRX]"
        else:
            role = "[PTX]"

        # Display channels list
        channels_list = ", ".join([str(i) for i in self.channels])
        channels = (
            f"channels=[{channels_list}] / last_channel={self.last_channel:4d}"
        )
        if self.__rssi is not None:
            return f"[{self.__rssi:4d} dBm] {role} {self.__address} {channels} {app_layer}"
        return f"{role} {self.__address} {channels} {app_layer}"


    def update_rssi(self, rssi: int =0):
        """Update device RSSI.

        :param  rssi: New RSSI value.
        :type   rssi: float
        """
        self.__rssi = rssi


    def update_channel(self, channel: int):
        """Update device channels in use.

        :param  channel: New channel value.
        :type   channel: int
        """
        self.__channels.append(channel)

    def set_applicative_layer(self, applicative_layer):
        """Update applicative layer.

        :param  applicative_layer: New applicative layer
        :type   applicative_layer:   str
        """
        if not self.__applicative_layer:
            self.__applicative_layer = applicative_layer


class CommunicatingDevicesDB:
    """Enhanced ShockBurst devices database.

    This class stores information about discovered devices.
    """

    def __init__(self):
        self.reset()


    def reset(self):
        """Remove database content.
        """
        self.__db = {}


    def find_device(self, address, role) -> CommunicatingDevice:
        """Find a device based on its address and role.

        :param      address: Enhanced ShockBurst address
        :type       address: str
        :param      role: Enhanced ShockBurst role
        :type       role: ESBRole
        :return:    Device if found, `None` otherwise.
        :rtype:     :class:`whad.esb.scanning.CommunicatingDevice`
        """
        if (address,role) in self.__db:
            return self.__db[(address,role)]
        return None


    def register_device(self, device: CommunicatingDevice):
        """Register or update a device.

        :param device: Device to register
        :type device: CommunicatingDevice
        """
        # Save it into our DB
        self.__db[(device.address,device.role)] = device


    def update_device(self, device: CommunicatingDevice, rssi: int, channel: int,
                      app_layer: str):
        """Update a device 

        :param device: Device to update
        :type device: CommunicatingDevice
        :param rssi: Device received signal strength indicator
        :type rssi: int
        :param channel: Channel on which the device has been last seen
        :type channel: int
        :param app_layer: Application layer for this device
        :type app_layer: str
        """
        # Update existing device RSSI and channel
        device.update_rssi(rssi)
        device.update_channel(channel)
        if app_layer is not None:
            device.set_applicative_layer(app_layer)


    def on_device_found(self, rssi, packet, filter_addr):
        """Device packet received.

        Parse the incoming packet and handle device appropriately.

        :param  rssi:           Received Signal Strength Indicator
        :type   rssi:           float
        :param  packet:         Received packet
        :type   packet:         :class:`scapy.packet.Packet`
        :param  filter_addr:    ESB address to filter
        :type   filter_addr:    str
        """
        address = ESBAddress(packet.metadata.address)
        if ESB_Payload_Hdr in packet:
            pdu = packet[ESB_Payload_Hdr:]
        else:
            pdu = ESB_Payload_Hdr()

        channel = packet.metadata.channel
        if ESB_Ack_Response in pdu or len(bytes(pdu)) == 0:
            role = ESBRole.PRX
        else:
            role = ESBRole.PTX

        applicative_layer = None

        payload = bytes(pdu)
        # Check if the payload is unifying
        if len(payload) >= 2 and payload[1] in UNIFYING_COMMAND_BYTES:
            checksum = 0x00
            for i in payload[:-1]:
                checksum = (checksum  - i) & 0xFF
            if checksum == payload[-1]:
                applicative_layer = "unifying"

        # If bd address does not match, don't report it
        if filter_addr is not None and filter_addr.lower() != str(address).lower():
            return None

        # No new device found for now
        new_device = None

        # Have we already discovered this device ?
        existing_device = self.find_device(str(address),role)
        if existing_device is None:
            # No, create and register new device
            new_device = CommunicatingDevice(rssi, str(address), role, applicative_layer, channel)
            self.register_device(new_device)
        else:
            # Yes, update existing device RSSI and channel
            self.update_device(existing_device, rssi, channel, applicative_layer)

        # Return discovered device
        return new_device



"""WHAD Protocol BLE prepared sequences messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from whad.protocol.ble.ble_pb2 import PrepareSequenceCmd
from whad.hub.message import pb_bind, PbFieldInt, PbFieldBytes, PbMessageWrapper, \
    PbFieldArray, Registry, HubMessage
from whad.hub.ble import BleDomain
from whad.hub.events import TriggeredEvt

class IPacketSequence:
    """Packet Sequence interface class
    """

    def add_packet(self, packet: bytes):
        """Add packet to the list of packets to send.
        """
        new_pending_pkt = self.message.ble.prepare.sequence.add()
        new_pending_pkt.packet = bytes(packet)

    def get_packet(self, index: int) -> bytes:
        """Get packet from the sequence.
        """
        if index >= 0 and index <= len(self.packets):
            return self.packets[index].packet
        
    def count_packets(self) -> int:
        """Return the number of packets in this prepare sequence message.
        """
        return len(self.packets)

@pb_bind(BleDomain, "prepare_manual", 1)
class PrepareSequenceManual(PbMessageWrapper, IPacketSequence):
    """BLE prepare sequence with manual trigger message class
    """

    sequence_id = PbFieldInt("ble.prepare.id")
    direction = PbFieldInt("ble.prepare.direction")
    packets = PbFieldArray("ble.prepare.sequence")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.message.ble.prepare.trigger.manual.CopyFrom(PrepareSequenceCmd.ManualTrigger())

@pb_bind(BleDomain, "prepare_connevt", 1)
class PrepareSequenceConnEvt(PbMessageWrapper, IPacketSequence):
    """BLE prepare sequence with connection event trigger message class
    """
    sequence_id = PbFieldInt("ble.prepare.id")
    direction = PbFieldInt("ble.prepare.direction")
    packets = PbFieldArray("ble.prepare.sequence")
    connection_event = PbFieldInt("ble.prepare.trigger.connection_event.connection_event")


@pb_bind(BleDomain, "prepare_pattern", 1)
class PrepareSequencePattern(PbMessageWrapper, IPacketSequence):
    """BLE prepare sequence with reception pattern trigger message class
    """
    sequence_id = PbFieldInt("ble.prepare.id")
    direction = PbFieldInt("ble.prepare.direction")
    packets = PbFieldArray("ble.prepare.sequence")
    pattern = PbFieldBytes("ble.prepare.trigger.reception.pattern")
    mask = PbFieldBytes("ble.prepare.trigger.reception.mask")
    offset = PbFieldInt("ble.prepare.trigger.reception.offset")

@pb_bind(BleDomain, "prepare", 1)
class PrepareSequence(Registry):

    def __init__(self, version: int):
        self.proto_version = version

    @staticmethod
    def parse(version: int, message: Message) -> HubMessage:
        """Parses a WHAD BLE PrepareSequence message as seen by protobuf
        """
        trigger_type = message.ble.prepare.trigger.WhichOneof("trigger")
        if trigger_type == 'manual':
            return PrepareSequenceManual.parse(version, message)
        elif trigger_type == 'connection_event':
            return PrepareSequenceConnEvt.parse(version, message)
        elif trigger_type == 'reception':
            return PrepareSequencePattern.parse(version, message)
        else:
            return None

@pb_bind(BleDomain, "triggered", 1)
class Triggered(PbMessageWrapper):
    """BLE prepare sequence triggered message class.

    This class implements WHAD's AbstractEvent interface and can be converted
    into a `TriggeredEvt` object.    
    """
    sequence_id = PbFieldInt("ble.triggered.id")


    def to_event(self):
        """Convert this message to an event.
        """
        return TriggeredEvt(
            sequence_id=self.sequence_id,
        )
    
    @staticmethod
    def from_event(event):
        return Triggered(
            sequence_id=event.sequence_id,
        )

@pb_bind(BleDomain, "trigger", 1)
class Trigger(PbMessageWrapper):
    """BLE prepare sequence trigger message class
    """
    sequence_id = PbFieldInt("ble.trigger.id")

@pb_bind(BleDomain, "delete_seq", 1)
class DeleteSequence(PbMessageWrapper):
    """BLE prepare sequence delete message class
    """
    sequence_id = PbFieldInt("ble.delete_seq.id")



"""LoRaWAN exceptions
"""

class ChannelNotFound(Exception):
    """This exception is raised when no channel can be found for
    uplink or downlink, from the current frequency plan.
    """
    def __init__(self):
        super().__init__()


class InvalidDataRate(Exception):
    """This exception is raised when a required datarate does not
    correspond to one of the datarates available in the channel
    plan.
    """
    def __init__(self):
        super().__init__()

class NotStartedException(Exception):
    def __init__(self):
        super().__init__()

class BadEuiFormat(Exception):
    def __init__(self):
        super().__init__()

class BadMICError(Exception):
    def __init__(self):
        super().__init__()

class MissingKeyError(Exception):
    def __init__(self, keyname):
        super().__init__()
        self.keyname = keyname

    def __repr__(self):
        return 'MissingKeyError(%s)' % self.keyname
    
class InvalidNodeRegistryError(Exception):
    def __init__(self, registry_path):
        self.registry_path = registry_path

    def __repr__(self):
        return 'InvalidNodeRegistryError(path="%s")' % self.registry_path


"""BLE characteristic write command handler
"""
from whad.cli.app import command
from whad.hub.ble.bdaddr import BDAddress
from whad.ble.profile.attribute import UUID, InvalidUUIDException
from whad.ble.stack.att.exceptions import AttError
from whad.ble.stack.gatt.exceptions import GattTimeoutException
from whad.ble.cli.central.helpers import show_att_error, create_central

# Expected parameters that must be passed to our program
# to use an already established connection
EXPECTED_BLE_PARAMS = [
    "initiator_bdaddr",
    "initiator_addrtype",
    "target_bdaddr",
    "target_addrtype",
    "conn_handle",
]

@command('write')
def write_handler(app, command_args):
    """write data to a a GATT attribute

    <ansicyan><b>write</b> <i>[UUID | handle] [hex [value] | value ]</i></ansicyan>

    Write data to the specified GATT attribute (identified by its handle) or to
    a characteristic value (identified by its UUID, if unique).

    Data can be provided hex-encoded if prefixed by "hex":

    > write 41 hex 41 42 43

    The command above will write 'ABC' to attribute identified by the handle 41.
    
    Data can also be provided as text:

    > write 41 ABC

    """
    if app.is_piped_interface():
        # Make sure we have all the required parameters
        for param in EXPECTED_BLE_PARAMS:
            if not hasattr(app.args, param):
                app.error("Source interface does not provide a BLE connection")

        # Create Central connector based on app configuration
        central, profile_loaded = create_central(app, piped=True)

        # If no connector returned, there was an error, simply exit.
        if central is None:
            return

        device = central.peripheral()

        # Read GATT characteristic
        perform_write(
            app,
            device,
            command_args,
            without_response=False,
            profile_loaded=profile_loaded
        )

    # We need to have an interface specified
    elif app.interface is not None and app.args.bdaddr is not None:

        # Make sure BD address is valid
        if not BDAddress.check(app.args.bdaddr):
            app.error("Invalid BD address: %s", app.args.bdaddr)
            return

        # Create Central connector based on app configuration
        central, profile_loaded = create_central(app, piped=False)

        # If no connector returned, there was an error, simply exit.
        if central is None:
            return

        # Start central
        central.start()

        # Connect to target device
        device = central.connect(app.args.bdaddr, random=app.args.random)
        if device is None:
            app.error("Cannot connect to %s, device does not respond.", app.args.bdaddr)
        else:
            # Perform write
            perform_write(
                app,
                device,
                command_args,
                without_response=False
            )

        # Disconnect
        device.disconnect()
        central.stop()
    else:
        app.error("You need to specify a target device with option --bdaddr.")


@command('writecmd')
def writecmd_handler(app, command_args):
    """write data to a a GATT attribute without (no response)

    <ansicyan><b>writecmd</b> <i>[UUID | handle] [hex [value] | value ]</i></ansicyan>

    Write data to the specified GATT attribute (identified by its handle) or to
    a characteristic value (identified by its UUID, if unique) without waiting
    for a response.

    Data can be provided hex-encoded if prefixed by "hex":

    > writecmd 41 hex 41 42 43

    The command above will write 'ABC' to attribute identified by the handle 41.
    
    Data can also be provided as text:

    > writecmd 41 ABC
    """
    if app.is_piped_interface():
        # Make sure we have all the required parameters
        for param in EXPECTED_BLE_PARAMS:
            if not hasattr(app.args, param):
                app.error("Source interface does not provide a BLE connection")

        # Create Central connector based on app configuration
        central, profile_loaded = create_central(app, piped=True)

        # If no connector returned, there was an error, simply exit.
        if central is None:
            return

        device = central.peripheral()

        # Read GATT characteristic
        perform_write(
            app,
            device,
            command_args,
            without_response=True,
            profile_loaded=profile_loaded
        )

        # Disconnect
        device.disconnect()
        central.stop()

    # We need to have an interface specified
    elif app.interface is not None and app.args.bdaddr is not None:

        # Make sure BD address is valid
        if not BDAddress.check(app.args.bdaddr):
            app.error(f"Invalid BD address: {app.args.bdaddr}")
            return

        # Create Central connector based on app configuration
        central, profile_loaded = create_central(app, piped=False)

        # If no connector returned, there was an error, simply exit.
        if central is None:
            return

        # Switch to central mode
        central.start()

        # Connect to target device
        device = central.connect(app.args.bdaddr, random=app.args.random)
        if device is None:
            app.error(f"Cannot connect to {app.args.bdaddr}, device does not respond.")
        else:
            # Perform write
            perform_write(
                app,
                device,
                command_args,
                without_response=True
            )

            # Disconnect
            device.disconnect()

        # Stop central role
        central.stop()

    else:
        app.error("You need to specify a target device with option --bdaddr.")


def perform_write(app, device, args, without_response=False, profile_loaded=False):
    """Perform attribute/handle characteristic
    """
    # parse target arguments
    if len(args) <2:
        app.error(("You must provide at least a characteristic value handle"
                   "or characteristic UUID, and a value to write."))
        return

    handle = None

    # Figure out what the handle is
    if args[0].lower().startswith('0x'):
        try:
            handle = int(args[0].lower(), 16)
        except ValueError:
            app.error("Wrong handle: %s", args[0])
            return
    else:
        try:
            handle = int(args[0])
        except ValueError:
            try:
                handle = UUID(args[0].replace('-',''))
            except InvalidUUIDException:
                app.error("Wrong UUID: %s", args[0])
                return

    # Do we have hex data ?
    if args[1].lower() == 'hex':
        # Decode hex data
        hex_data = ''.join(args[2:])
        try:
            char_value = bytes.fromhex(hex_data.replace("\t", ""))
        except ValueError:
            app.error("Provided hex value contains non-hex characters.")
            return
    else:
        char_value = args[1]

    if not isinstance(char_value, bytes):
        char_value = bytes(char_value,'utf-8')

    # Perform ATT write by handle
    if not isinstance(handle, UUID):
        try:
            if without_response:
                device.write_command(handle, char_value)
            else:
                device.write(handle, char_value)
        except AttError as att_err:
            show_att_error(app, att_err)
        except GattTimeoutException:
            app.error('GATT timeout while writing.')
    else:
        if not profile_loaded:
            # Perform discovery if required
            device.discover()

        # Search characteristic from its UUID
        target_charac = device.find_characteristic_by_uuid(handle)
        if target_charac is not None:
            try:
                if without_response:
                    target_charac.write(char_value, without_response=True)
                else:
                    target_charac.value = char_value
            except AttError as att_err:
                show_att_error(app, att_err)
            except GattTimeoutException:
                app.error('GATT timeout while writing.')
        else:
            app.error("No characteristic found with UUID %s", handle)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: whad/protocol/whad.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from whad.protocol import device_pb2 as whad_dot_protocol_dot_device__pb2
from whad.protocol import generic_pb2 as whad_dot_protocol_dot_generic__pb2
from whad.protocol.ble import ble_pb2 as whad_dot_protocol_dot_ble_dot_ble__pb2
from whad.protocol.dot15d4 import dot15d4_pb2 as whad_dot_protocol_dot_dot15d4_dot_dot15d4__pb2
from whad.protocol.esb import esb_pb2 as whad_dot_protocol_dot_esb_dot_esb__pb2
from whad.protocol.unifying import unifying_pb2 as whad_dot_protocol_dot_unifying_dot_unifying__pb2
from whad.protocol.phy import phy_pb2 as whad_dot_protocol_dot_phy_dot_phy__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18whad/protocol/whad.proto\x1a\x1awhad/protocol/device.proto\x1a\x1bwhad/protocol/generic.proto\x1a\x1bwhad/protocol/ble/ble.proto\x1a#whad/protocol/dot15d4/dot15d4.proto\x1a\x1bwhad/protocol/esb/esb.proto\x1a%whad/protocol/unifying/unifying.proto\x1a\x1bwhad/protocol/phy/phy.proto\"\x81\x02\n\x07Message\x12#\n\x07generic\x18\x01 \x01(\x0b\x32\x10.generic.MessageH\x00\x12\'\n\tdiscovery\x18\x02 \x01(\x0b\x32\x12.discovery.MessageH\x00\x12\x1b\n\x03\x62le\x18\x03 \x01(\x0b\x32\x0c.ble.MessageH\x00\x12#\n\x07\x64ot15d4\x18\x04 \x01(\x0b\x32\x10.dot15d4.MessageH\x00\x12\x1b\n\x03\x65sb\x18\x05 \x01(\x0b\x32\x0c.esb.MessageH\x00\x12%\n\x08unifying\x18\x06 \x01(\x0b\x32\x11.unifying.MessageH\x00\x12\x1b\n\x03phy\x18\x07 \x01(\x0b\x32\x0c.phy.MessageH\x00\x42\x05\n\x03msgb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'whad.protocol.whad_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MESSAGE._serialized_start=249
  _MESSAGE._serialized_end=506
# @@protoc_insertion_point(module_scope)



class ModulationScheme:
    def __init__(self, symbols):
        self.symbols = symbols

    def __repr__(self):
        return self.__class__.__name__+"(symbols=({}))".format(",".join(self.symbols))

class ASKModulationScheme(ModulationScheme):
    def __init__(self, symbols, on_off_keying=False):
        self.on_off_keying = on_off_keying
        super().__init__(symbols)

class OOKModulationScheme(ASKModulationScheme):
    def __init__(self):
        super().__init__(("0", "1"), on_off_keying=True)

class FSKModulationScheme(ModulationScheme):
    def __init__(self, symbols, deviation, gaussian_filter=False):
        self.deviation = deviation
        self.gaussian_filter = gaussian_filter
        super().__init__(symbols)

class QFSKModulationScheme(FSKModulationScheme):
    def __init__(self, deviation):
        super().__init__(("00", "01", "11", "10"), deviation=deviation)


class GFSKModulationScheme(FSKModulationScheme):
    def __init__(self, deviation):
        super().__init__(("0", "1"), deviation, gaussian_filter=True)

class PSKModulationScheme(ModulationScheme):
    def __init__(self, symbols, offset_mode=False):
        self.offset_mode = offset_mode
        super().__init__(symbols)

class BPSKModulationScheme(PSKModulationScheme):
    def __init__(self, offset_mode=False):
        super().__init__(("0", "1"), offset_mode)


class QPSKModulationScheme(PSKModulationScheme):
    def __init__(self, offset_mode=False):
        super().__init__(("00", "01", "10", "11"), offset_mode)

class OQPSKModulationScheme(QPSKModulationScheme):
    def __init__(self):
        super().__init__(offset_mode=True)

class PhysicalLayer:
    def __init__(self, modulation, datarate, endianness, synchronization_word, frequency_range, maximum_packet_size, configuration=None, scapy_layer=None, channel_to_frequency_function=None, frequency_to_channel_function=None, format_address_function=None, integrity_function=None, decoding_function=None, encoding_function=None):
        self.modulation = modulation
        self.datarate = datarate
        self.endianness = endianness
        self.synchronization_word = synchronization_word
        self.frequency_range = frequency_range
        self.maximum_packet_size = maximum_packet_size
        self.configuration = configuration
        self.scapy_layer = scapy_layer
        self.channel_to_frequency = channel_to_frequency_function
        self.frequency_to_channel = frequency_to_channel_function
        self.format_address = format_address_function
        self.integrity = integrity_function
        self.decoding = decoding_function
        self.encoding = encoding_function



from whad.zigbee.connector import Zigbee, Sniffer, EndDevice, Coordinator
from whad.zigbee.utils.phy import PHYS
__all__ = [
    'Zigbee',
    'Sniffer',
    'EndDevice',
    'Coordinator', 
    'PHYS'
]



"""802.15.4 Medium Access Control Layer (MAC) exceptions
"""

class MACTimeoutException(Exception):
    """This exception is raised when a timeout occurs at MAC level.
    """
    def __init__(self):
        super().__init__()

class MACAssociationFailure(Exception):
    """This exception is raised when association fails at MAC level.
    """
    def __init__(self, reason):
        super().__init__()
        self.reason = reason



import logging
from typing import Generator
from time import time

from scapy.packet import Packet

from whad.rf4ce.connector import RF4CE
from whad.rf4ce.sniffing import SnifferConfiguration
from whad.rf4ce.crypto import RF4CEDecryptor, RF4CEKeyDerivation
from whad.rf4ce.utils.adpcm import ADPCM
from whad.rf4ce.exceptions import MissingCryptographicMaterial
from whad.exceptions import UnsupportedCapability
from whad.helpers import message_filter, is_message_type
from whad.rf4ce.sniffing import KeyExtractedEvent
from whad.common.sniffing import EventsManager
from whad.hub.dot15d4 import RawPduReceived, PduReceived
from whad.hub.message import AbstractPacket
from whad.exceptions import WhadDeviceDisconnected


logger = logging.getLogger(__name__)

class Sniffer(RF4CE, EventsManager):
    """
    RF4CE Sniffer interface for compatible WHAD device.
    """

    def __init__(self, device, configuration=SnifferConfiguration()):
        RF4CE.__init__(self, device)
        EventsManager.__init__(self)


        self.__configuration = configuration
        self.__decryptor = RF4CEDecryptor()
        self.__key_derivation = RF4CEKeyDerivation()
        self.__audio_stream = ADPCM()

        # Check if device can perform sniffing
        if not self.can_sniff():
            raise UnsupportedCapability("Sniff")

    def _enable_sniffing(self):
        for key in self.__configuration.keys:
            self.__decryptor.add_key(key)
        for address in self.__configuration.addresses:
            self.__decryptor.add_address(address)
        if (
            self.__configuration.audio and
            self.__configuration.audio_file is not None and
            self.__configuration.audio_file
        ):
            self.__audio_stream.output_filename = self.__configuration.audio_file
        self.sniff_rf4ce(channel=self.__configuration.channel)

    def add_key(self, key):
        self.__configuration.keys.append(key)

    def clear_keys(self):
        self.__configuration.keys = []

    def add_address(self, key):
        self.__configuration.addresses.append(address)

    def clear_addresses(self):
        self.__configuration.addresses = []

    @property
    def decrypt(self):
        return self.__configuration.decrypt

    @decrypt.setter
    def decrypt(self, decrypt):
        self.__configuration.decrypt = decrypt


    @property
    def channel(self):
        return self.__configuration.channel

    @channel.setter
    def channel(self, channel=15):
        self.stop()
        self.__configuration.channel = channel
        self._enable_sniffing()


    @property
    def configuration(self):
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration):
        #self.stop()
        self.__configuration = new_configuration
        self._enable_sniffing()

    def process_packet(self, packet):

        if self.__configuration.pairing:
            self.__key_derivation.process_packet(packet)
            if self.__key_derivation.key is not None:
                logger.info("[i] New key extracted: ", self.__key_derivation.key.hex())
                self.trigger_event(KeyExtractedEvent(self.__key_derivation.key))
                self.__decryptor.add_key(self.__key_derivation.key)
                self.__key_derivation.reset()

        if (
            hasattr(packet, "fcf_destaddrmode") and
            packet.fcf_destaddrmode == 3
        ):
            self.__decryptor.add_address(packet.dest_addr)

        if (
            hasattr(packet, "fcf_srcaddrmode") and
            packet.fcf_srcaddrmode == 3
        ):
            self.__decryptor.add_address(packet.src_addr)

        if (
            hasattr(packet, "security_enabled") and
            packet.security_enabled == 1 and
            self.__configuration.decrypt
        ):
            try:
                decrypted, success = self.__decryptor.attempt_to_decrypt(packet)
                if success:
                    metadata = packet.metadata
                    packet = decrypted
                    packet.metadata = metadata
                    if self.__configuration.audio:
                        self.__audio_stream.process_packet(packet)

            except MissingCryptographicMaterial:
                pass
        return packet

    def sniff(self, timeout: float = None) -> Generator[Packet, None, None]:
        """Sniff RF4CE packets out of thin air.

        :param timeout: Number of seconds after which sniffing will stop. Wait
                        forever if set to `None`.
        :type timeout: float
        """
        if self.support_raw_pdu():
            message_type = RawPduReceived
        else:
            message_type = PduReceived

        try:
            for message in super().sniff(messages=(message_type), timeout=timeout):
                if message is not None and issubclass(message, AbstractPacket):
                    packet = message.to_packet()
                    if packet is not None:
                        packet = self.process_packet(packet)
                        self.monitor_packet_rx(packet)
                        yield packet
        except WhadDeviceDisconnected:
            return



from enum import IntEnum

class ClickType(IntEnum):
    NONE = 0
    LEFT = 1
    RIGHT = 2
    LEFT_RIGHT = 3
    MIDDLE = 4
    LEFT_MIDDLE = 5
    RIGHT_MIDDLE = 6

    def __str__(self):
        if self == ClickType.NONE:
            return ""
        elif self == ClickType.LEFT:
            return "L"
        elif self == ClickType.RIGHT:
            return "R"
        elif self == ClickType.LEFT_RIGHT:
            return "LR"
        elif self == ClickType.MIDDLE:
            return "M"
        elif self == ClickType.LEFT_MIDDLE:
            return "LM"
        elif self == ClickType.RIGHT_MIDDLE:
            return "RM"

class UnifyingRole(IntEnum):
    DONGLE = 0
    MOUSE = 1
    KEYBOARD = 2


class MultimediaKey(IntEnum):
    VOLUME_DOWN = 0xEA
    VOLUME_UP = 0xE9
    VOLUME_TOGGLE = 0xE2



from whad.zigbee.stack.apl.zcl import ZCLCluster, ZCLClientCluster, ZCLClusterConfiguration
from whad.scapy.layers.zll import ZigbeeZLLCommissioningCluster, ZLLScanRequest, ZLLIdentifyRequest, \
    ZLLDeviceInformationRequest, ZLLResetToFactoryNewRequest, ZLLNetworkJoinRouterRequest, ZLLNetworkStartRequest
from whad.dot15d4.stack.mac.constants import MACAddressMode
from whad.zigbee.profile.network import Network
from whad.zigbee.crypto import TouchlinkKeyManager

from struct import pack, unpack
from time import time
from random import randint

class ZCLTouchLinkClient(ZCLClientCluster):

    def __init__(self):
        self.transaction_id = None
        super().__init__(
            cluster_id=0x1000,
            default_configuration=ZCLClusterConfiguration(
                destination_address=0xFFFF,
                destination_pan_id =0xFFFF,
                interpan=True,
                disable_default_response=True
            )
        )

    def scan(self, channels = [11, 15, 20, 25], link_initiator=True, address_assignment=True, factory_new=True):
        """
        Perform a touchlink scan and return detected networks.
        """
        # Select the channel page 0
        mac_layer = self.application.manager.get_layer('mac')
        nwk_layer = self.application.manager.get_layer('nwk')

        mac_layer.set_channel_page(0)

        for channel in channels:
            # Go to touchlink channel
            mac_layer.set_channel(channel)
            # Send a scan request and wait for response
            scan_response = self.scan_request(
                link_initiator = link_initiator,
                address_assignment = address_assignment,
                factory_new = factory_new
            )
            if scan_response is not None:
                pdu, source = scan_response
                print("Source", source)
                pdu.show()
                a = self.reset_to_factory_new(destination_address=source, transaction_id=pdu.inter_pan_transaction_id)
                print("a", a)
                '''
                network_address = randint(2, 0xFFF0)
                resp = self.network_join_router_request(
                                                transaction_id=pdu.inter_pan_transaction_id,
                                                response_id=pdu.response_id,
                                                destination_address=source,
                                                logical_channel=0,
                                                pan_id=pdu.pan_id,
                                                extended_pan_id=pdu.pan_id_ext,
                                                network_address=network_address
                )
                '''

                return network_address


    @ZCLCluster.command_generate(0x00, "ScanRequest")
    def scan_request(self, transaction_id=None, link_initiator=True, address_assignment=True, factory_new=True):
        """
        Trigger a scan request.
        """
        # Get a new transaction ID if needed.
        if transaction_id is None:
            transaction_id = randint(0, 0xFFFFFFFF)
            self.transaction_id = transaction_id

        # Get the local node descriptor
        node_descriptor = self.application.manager.get_application_by_name("zdo").configuration.get("configNodeDescriptor")
        # Build and send a Scan Request
        command = ZLLScanRequest(
            inter_pan_transaction_id=transaction_id,
            rx_on_when_idle=int(node_descriptor.receiver_on_when_idle),
            logical_type=int(node_descriptor.logical_type),
            link_initiator=int(link_initiator),
            address_assignment=int(address_assignment),
            factory_new=int(factory_new)
        )

        self.send_command(command)
        # Return the received response (if any)
        return self.wait_response()

    @ZCLCluster.command_receive(0x01, "ScanResponse")
    def on_scan_response(self, command, source_address):
        """
        Process a scan response.
        """
        self.push_response(
            (command, source_address),
            command.inter_pan_transaction_id
        )




    @ZCLCluster.command_generate(0x02, "DeviceInformationRequest")
    def device_information_request(self,transaction_id=None, start_index=0, destination_address=None):
        """
        Performs a Device Information request.
        """
        # Get a new transaction ID if needed.
        if transaction_id is None:
            transaction_id = randint(0, 0xFFFFFFFF)
            self.transaction_id = transaction_id

        print(transaction_id)
        self.configure(
            destination_address=destination_address,
            destination_address_mode=MACAddressMode.EXTENDED,
            interpan=True,
            acknowledged_transmission=True,
            disable_default_response=True
        )

        command = ZLLDeviceInformationRequest(inter_pan_transaction_id=transaction_id, start_index=start_index)
        command.show()
        self.send_command(command)
        return self.wait_response()

    @ZCLCluster.command_generate(0x06, "IdentifyRequest")
    def identify_request(self,transaction_id=None, identify_duration=1, destination_address=None):
        """
        Performs an Identify request.
        """
        if transaction_id is None:
            transaction_id = randint(0, 0xFFFFFFFF)
            self.transaction_id = transaction_id

        self.configure(
            destination_address=destination_address,
            destination_address_mode=MACAddressMode.EXTENDED,
            interpan=True,
            acknowledged_transmission=True,
            disable_default_response=True
        )

        command = ZLLIdentifyRequest(inter_pan_transaction_id=transaction_id, identify_duration=identify_duration)
        self.send_command(command)

    @ZCLCluster.command_generate(0x07, "ResetToFactoryNew")
    def reset_to_factory_new(self,transaction_id=None, destination_address=None):
        """
        Performs a reset to factory new.
        """
        if transaction_id is None:
            transaction_id = randint(0, 0xFFFFFFFF)
            self.transaction_id = transaction_id

        self.configure(
            destination_address=destination_address,
            destination_address_mode=MACAddressMode.EXTENDED,
            interpan=True,
            acknowledged_transmission=True,
            disable_default_response=True
        )
        command = ZLLResetToFactoryNewRequest(inter_pan_transaction_id=transaction_id)
        self.send_command(command)

    @ZCLCluster.command_generate(0x12, "NetworkJoinRouterRequest")
    def network_join_router_request(
                                    self,
                                    transaction_id=None,
                                    response_id=None,
                                    destination_address=None,
                                    key_index=4,
                                    network_update_id=0,
                                    key=bytes.fromhex("01020102030403040506050607080708")[::-1],
                                    logical_channel=25,
                                    pan_id=0x1234,
                                    extended_pan_id=0x0102030405060708,
                                    network_address=0x0001
    ):
        """
        Performs a network join router request.
        """
        # Get a new transaction ID if needed.
        if transaction_id is None:
            transaction_id = randint(0, 0xFFFFFFFF)
            self.transaction_id = transaction_id

        # Configure the transmission parameters
        self.configure(
            destination_address=destination_address,
            destination_address_mode=MACAddressMode.EXTENDED,
            interpan=True,
            acknowledged_transmission=True,
            disable_default_response=True
        )

        # Encrypt the key
        encrypted_key = TouchlinkKeyManager(
            unencrypted_key=key,
            response_id=response_id,
            transaction_id=transaction_id,
            key_index=key_index
        ).encrypted_key

        command = ZLLNetworkJoinRouterRequest(
            inter_pan_transaction_id=transaction_id,
            pan_id_ext=extended_pan_id,
            key_index=key_index,
            encrypted_network_key=encrypted_key,
            network_update_id=network_update_id,
            channel=logical_channel,
            pan_id=pan_id,
            network_address=network_address
        )

        self.send_command(command)
        return self.wait_response()

    @ZCLCluster.command_receive(0x13, "NetworkJoinRouterResponse")
    def on_network_join_router_response(self, command, source_address):
        """
        Process a network join router response.
        """
        command.show()
        self.push_response(
            (command, source_address),
            command.inter_pan_transaction_id
        )


    @ZCLCluster.command_generate(0x10, "NetworkStartRequest")
    def network_start_router_request(
                                    self,
                                    transaction_id=None,
                                    response_id=None,
                                    destination_address=None,
                                    key_index=4,
                                    key=bytes.fromhex("11223344556677881122334455667788"),
                                    logical_channel=25,
                                    pan_id=0x1234,
                                    extended_pan_id=0x0102030405060708,
                                    network_address=0x0001,
                                    initiator_network_address=0x0002,
                                    initiator_ieee_address=None
    ):
        """
        Performs a network start request.
        """
        # Get a new transaction ID if needed.
        if transaction_id is None:
            transaction_id = randint(0, 0xFFFFFFFF)
            self.transaction_id = transaction_id

        if initiator_ieee_address is None:
            initiator_ieee_address = unpack('<Q',
                pack(">Q",
                    self.application.manager.get_layer("nwk").database.get("nwkIeeeAddress")
                )
            )[0]

        # Configure the transmission parameters
        self.configure(
            destination_address=destination_address,
            destination_address_mode=MACAddressMode.EXTENDED,
            interpan=True,
            acknowledged_transmission=True,
            disable_default_response=True
        )

        # Encrypt the key
        encrypted_key = TouchlinkKeyManager(
            unencrypted_key=key,
            response_id=response_id,
            transaction_id=transaction_id,
            key_index=key_index
        ).encrypted_key

        command = ZLLNetworkStartRequest(
            inter_pan_transaction_id=transaction_id,
            pan_id_ext=extended_pan_id,
            key_index=key_index,
            encrypted_network_key=encrypted_key,
            channel=logical_channel,
            pan_id=pan_id,
            network_address=network_address,
            initiator_ieee_address=initiator_ieee_address,
            initiator_network_address=initiator_network_address,
        )

        self.send_command(command)
        return self.wait_response()


    @ZCLCluster.command_receive(0x03, "DeviceInformationResponse")
    def on_device_information_response(self, command, source_address):
        """
        Process a device information response.
        """
        self.push_response(
            (command, source_address),
            command.inter_pan_transaction_id
        )

    @ZCLCluster.command_receive(0x11, "NetworkStartResponse")
    def on_network_start_response(self, command, source_address):
        """
        Process a network start response.
        """
        self.push_response(
            (command, source_address),
            command.inter_pan_transaction_id
        )

    def push_response(self, response, transaction_id):
        """
        Adding a response to the pending responses.
        """
        # Check if a pending response is expected
        if transaction_id in self.pending_responses:
            # Add the response
            self.pending_responses[transaction_id] = response

    def wait_response(self, transaction=None, timeout=1):
        """
        Wait for the response associated to a specific transaction (default: transaction id).
        """
        # if transaction not provided, use the last transaction by default
        if transaction is None:
            transaction = self.transaction_id

        # Clear the entry associated to the provided transaction in pending responses dictionnary
        self.pending_responses[transaction] = None

        start = time()
        while self.pending_responses[transaction] is None and time() - start < timeout:
            pass
        # If we got a response return it, otherwise a timeout occurred and we return None
        if self.pending_responses[transaction] is not None:
            return_value = self.pending_responses[transaction]
        else:
            return_value = None
        del self.pending_responses[transaction]
        return return_value



"""
Bluetooth LE Stack L2CAP manager
"""
import logging
from struct import unpack, pack

from scapy.layers.bluetooth import L2CAP_Hdr, ATT_Hdr, SM_Hdr, L2CAP_CmdHdr, \
    L2CAP_Connection_Parameter_Update_Request, L2CAP_Connection_Parameter_Update_Response

from whad.ble.stack.att import ATTLayer
from whad.ble.stack.smp import SMPLayer

from whad.common.stack import Layer, alias, source, state, LayerState, ContextualLayer

logger = logging.getLogger(__name__)

@alias('l2cap')
class L2CAPLayer(ContextualLayer):
    """L2CAP layer implementation
    """

    def configure(self, options=None):
        """Layer configuration
        """
        # Initialize state
        self.state.conn_handle = None
        self.state.local_mtu = 23
        self.state.remote_mtu = 23
        self.state.fifo = None
        self.state.expected_len = 0

    def set_conn_handle(self, conn_handle: int):
        '''Save current connection handle
        '''
        self.state.conn_handle = conn_handle

    def get_conn_handle(self) -> int:
        """Retrieve current connection handle
        """
        return self.state.conn_handle

    def set_local_mtu(self, mtu: int):
        """Set local MTU
        """
        logger.debug("local MTU changed to %d for conn_handle %d", mtu, self.state.conn_handle)
        self.state.local_mtu = mtu

    def set_remote_mtu(self, mtu: int):
        '''Set remote MTU
        '''
        logger.debug("remote MTU changed to %d for conn_handle %d", mtu, self.state.conn_handle)
        self.state.remote_mtu = mtu

        # Update local MTU if lower than remote MTU
        if self.state.local_mtu < self.state.remote_mtu:
            logger.debug("local MTU changed to %d for conn_handle %d",
                         self.state.remote_mtu, self.state.conn_handle)
            self.state.local_mtu = self.state.remote_mtu

    def get_local_mtu(self) -> int:
        """Retrieve local MTU
        """
        return self.state.local_mtu

    @source('ll')
    def on_data_received(self, l2cap_data, fragment=False):
        """Handles incoming L2CAP data"""
        if fragment and self.state.fifo is not None:
            logger.debug("[l2cap] Received a L2CAP fragment of %d bytes", len(l2cap_data))
            self.state.fifo += l2cap_data
            logger.debug("[l2cap] L2CAP packet size so far: %d", len(self.state.fifo))

            if len(self.state.fifo) >= self.state.expected_len:
                # We have received a complete L2CAP packet, process it
                logger.debug("[l2cap] Received a complete L2CAP packet, process it")
                self.on_l2cap_packet(L2CAP_Hdr(self.state.fifo[:self.state.expected_len]))
                self.state.fifo = None

        elif len(l2cap_data) >= 2:
            # Start of L2CAP or complete L2CAP message
            self.state.fifo = l2cap_data
            logger.debug("[l2cap] received start of fragmented or complete message")

            # Check if we have a complete L2CAP message
            self.state.expected_len = unpack('<H', self.state.fifo[:2])[0] + 4
            logger.debug("[l2cap] expected l2cap length: %d", self.state.expected_len)
            logger.debug("[l2cap] actual l2cap length: %d", len(self.state.fifo))
            if len(self.state.fifo) >= self.state.expected_len:
                # We have received a complete L2CAP packet, process it
                logger.debug("[l2cap] Received a complete L2CAP packet, process it")
                self.on_l2cap_packet(L2CAP_Hdr(self.state.fifo[:self.state.expected_len]))
                self.state.fifo = None


    def on_l2cap_packet(self, packet):
        """Process incoming L2CAP packets.
        """
        if ATT_Hdr in packet:
            self.send('att', packet.getlayer(ATT_Hdr))
        elif L2CAP_CmdHdr in packet:
            self.on_cmd_packet(packet.getlayer(L2CAP_CmdHdr))
        elif SM_Hdr in packet:
            self.send('smp', packet.getlayer(SM_Hdr))

    def on_cmd_packet(self, packet):
        """Handle L2CAP Connection Parameter Update Requests.

        This command is rejected by default.
        """
        if L2CAP_Connection_Parameter_Update_Request in packet:
            logger.debug("[l2cap] Received a L2CAP Connection Parameter Update Request, rejecting")
            # Reject it
            self.send("ll", L2CAP_Hdr()/L2CAP_CmdHdr(
                id=packet[L2CAP_CmdHdr].id
            )/L2CAP_Connection_Parameter_Update_Response(move_result=1))

    def get_fragments(self, data: bytes):
        """Split data into fragments based on MTU
        """
        packets=[]
        # If data is bigger than MTU-1, then split
        if len(data) > self.state.remote_mtu:
            nb_packets = int(len(data)/(self.state.remote_mtu - 1))
            if nb_packets * (self.state.remote_mtu - 1) < len(data):
                nb_packets += 1
            raw_data = bytes(data)
            for i in range(nb_packets):
                packets.append(
                    raw_data[i*(self.state.remote_mtu - 1):(i+1)*(self.state.remote_mtu - 1)]
                )
        else:
            nb_packets = 1
            packets = [
                data
            ]

        return packets

    @source('att')
    def on_att_packet_recv(self, data, channel='attribute'):
        """Process incoming packets from ATT that must be forwarded to link layer.
        """
        packets = self.get_fragments(data)

        # Send packets
        for i, pkt in enumerate(packets):
            # First packet is sent with no fragment flag
            if i == 0:
                # Send packet to link layer
                self.send('ll', L2CAP_Hdr(cid=channel, len=len(data))/pkt, fragment=False)
            else:
                # Send packet to link layer
                self.send('ll', pkt, fragment=True)


    @source('smp')
    def on_smp_packet_recv(self, data, channel=0x06):
        """Process incoming packets from ATT that must be forwarded to link layer.
        """
        # Note: we are forced to hardcode the channel in parameters here, because
        # scapy does not provide a complete dictionary mapping 0x06 to 'smp'.

        packets = self.get_fragments(data)

        # Send packets
        for i, pkt in enumerate(packets):
            # First packet is sent with no fragment flag
            if i == 0:
                # Send packet to link layer
                self.send('ll', L2CAP_Hdr(cid=channel, len=len(data))/pkt, fragment=False)
            else:
                # Send packet to link layer
                self.send('ll', pkt, fragment=True)

    @source("att", tag="ATT_MTU")
    def on_att_mtu_updated(self, mtu: int):
        """Notify link-layer that MTU has changed for this connection
        """
        self.send('ll', mtu, tag="ATT_MTU")

L2CAPLayer.add(ATTLayer)
L2CAPLayer.add(SMPLayer)



"""
This module provides the :class:`whad.zigbee.connector.enddevice.EndDevice`
connector that allows the creation of a ZigBee end device.

This connector is able to discover networks and based on a discovered network
to join it (if permitted) and act as an End Device in this network.
"""
from typing import List, Union
from scapy.packet import Packet

from whad.device import Device
from whad.dot15d4.stack import Dot15d4Stack
from whad.dot15d4.stack.mac import MACManager
from whad.exceptions import UnsupportedCapability

from .base import Zigbee
from ..stack.nwk import NWKManager
from ..stack.apl.constants import LogicalDeviceType
from ..stack.apl.application import ApplicationObject
from ..stack.nwk.network import ZigbeeNetwork

class EndDevice(Zigbee):
    """
    Zigbee End Device interface for compatible WHAD device.
    """

    def __init__(self, device: Device, applications: List[ApplicationObject] = None):
        """ZigBee End Device connector initialization.

        :param device: WHAD device
        :type device: :class:`whad.device.WhadDevice`
        :param applications: list of application objects to use with this end device
        :type applications: list
        """
        super().__init__(device)

        if not self.can_be_end_device():
            raise UnsupportedCapability("EndDevice")

        # Stack initialization
        MACManager.add(NWKManager)
        self.__stack = Dot15d4Stack(self)

        # Channel initialization
        self.__channel = 11
        self.__channel_page = 0

        self.enable_reception()
        zdo = self.__stack.get_layer('apl').get_application_by_name("zdo")
        zdo.configuration.get("configNodeDescriptor").logical_type = LogicalDeviceType.END_DEVICE
        self.__stack.get_layer('apl').initialize()
        self._init_applications(applications)

    def _init_applications(self, applications: List[ApplicationObject]):
        """Initialize ZigBee application objects attached to the end device.
        If no application is defined, add a default ZigBee Cluster Library application
        object.

        :param applications: list of application objects to attach to the end device
        :type applications: list
        """
        if applications is None:
            # If no application provided, attach a default ZCL application on endpoint 1
            app = ApplicationObject("zcl_app", 0x0104, 0x0100, device_version=0,
                                    input_clusters=[], output_clusters=[])
            self.__stack.get_layer('apl').attach_application(app, endpoint=1)

        else:
            for app in applications:
                endpoint = 1
                self.__stack.get_layer('apl').attach_application(app, endpoint=endpoint)
                endpoint += 1

    def discover_networks(self) -> List[ZigbeeNetwork]:
        """Discover ZigBee networks.

        :return: list of discovered ZigBee networks
        :rtype: list
        """
        return self.__stack.get_layer('apl').get_application_by_name("zdo").network_manager.discover_networks()

    @property
    def stack(self):
        """Current ZigBee stack
        """
        return self.__stack

    def enable_reception(self):
        """Enable reception mode on the currently selected channel (enable
        device mode).
        """
        self.set_end_device_mode(channel=self.__channel)

    def set_channel(self, channel: int = 11):
        """Set ZigBee channel to use for the end device.

        :param channel: channel to use
        :type channel: int
        """
        self.__channel = channel
        self.enable_reception()

    def perform_ed_scan(self, channel: int = 11) -> bool:
        """Perform an energy detection scan on the specified channel.

        :param channel: channel on which the detection will be performed
        :type channel: int
        :return: ``True`` on success, ``False`` otherwise.
        :rtype: bool
        """
        if not self.can_perform_ed_scan():
            raise UnsupportedCapability("EnergyDetection")
        self.__channel = channel
        return super().perform_ed_scan(channel)

    def set_channel_page(self, page: int = 0):
        """Set ZigBee channel page for the end device

        :param page: channel page to use
        :type page: int
        """
        if page != 0:
            raise UnsupportedCapability("ChannelPageSelection")
        else:
            self.__channel_page = page

    def get_channel(self) -> int:
        """Retrieve the currently selected channel.

        :return: current ZigBee channel number
        :rtype: int
        """
        return self.__channel

    def get_channel_page(self) -> int:
        """Retrieve the current channel page

        :return: current channel page value
        :rtype: int
        """
        return self.__channel_page


    def send(self, pdu: Union[Packet, bytes], channel: int = None):
        """Send a PDU to the associated network, if any.

        :param pdu: pdu to send
        :type pdu: :class:`scapy.packet.Packet`
        :type pdu: bytes
        :param channel: channel on which to send the pdu
        :type channel: int, optional
        """
        # Forward to Dot15d4 (override channel parameter)
        super().send(pdu, channel=self.__channel)

    def on_pdu(self, packet: Packet):
        """ZigBee received PDU handler.

        :param packet: received packet (PDU)
        :type packet: :class:`scapy.packet.Packet`
        """
        if (
            hasattr(packet,"metadata") and
            hasattr(packet.metadata, "is_fcs_valid") and
            not packet.metadata.is_fcs_valid
        ):
            return

        self.__stack.on_pdu(packet)

    def on_ed_sample(self, timestamp: int, sample: int):
        """Process receive energy detection sample.

        :param timestamp: timestamp in microseconds
        :type timestamp: int
        :param sample: sample measured
        :type sample: int
        """
        self.__stack.on_ed_sample(timestamp, sample)


"""RF4CE Application Layer (APL) exceptions
"""

class APLTimeoutException(Exception):
    """This exception is raised when a timeout occurs at APL level.
    """
    def __init__(self):
        super().__init__()



"""Basic WHAD message list replay Mock device.

This mock device takes a list of WHAD messages to send with an optional delay,
and will send them to the connector at the specified pace, without interaction.
"""
import logging
from time import sleep

from scapy.all import Packet, rdpcap

# Load WHAD dedicated layers
from whad.scapy.layers import *
from whad.hub.discovery import Domain

from whad.exceptions import WhadDeviceDisconnected
from .base import MockDevice


logger = logging.getLogger(__name__)

class ReplayMock(MockDevice):
    """Replay a series of WHAD messages and wait a given time between each message.
    """

    def __init__(self, *args, messages: list = None, delay:float = None,
                 domain: str = None, **kwargs):
        """Constructor."""
        super().__init__(*args, **kwargs)
        if isinstance(delay, int):
            delay = float(delay)
        elif not isinstance(delay, float):
            logger.warning("[%s] specified delay is not a floating-point number, assuming no delay",
                           self.interface)

        self.__delay = delay if isinstance(delay, float) else None
        self.__messages = [] if messages is None else messages

    @property
    def interface(self) -> str:
        """Mock interface alias for debugging."""
        return f"mock-replay{self.index}"

    def on_interface_message(self):
        """Each time this method is called, we wait `ifs` seconds and report
        a packet. If no more packet is available, we report this device as
        disconnected.
        """
        if len(self.__messages) > 0:
            if self.__delay is not None:
                sleep(self.__delay)
            return self.__messages.pop(0)
        else:
            raise WhadDeviceDisconnected()



"""WHAD Protocol Discovery info messages abstraction layer.
"""
from whad.hub.message import pb_bind, PbFieldInt, PbMessageWrapper
from whad.hub.discovery import Discovery

@pb_bind(Discovery, 'domain_query', 1)
class DomainInfoQuery(PbMessageWrapper):
    """Device info domain query message class
    """
    domain = PbFieldInt('discovery.domain_query.domain')

@pb_bind(Discovery, 'domain_resp', 1)
class DomainInfoQueryResp(PbMessageWrapper):
    """Device info domain query response message class
    """
    domain = PbFieldInt('discovery.domain_resp.domain')
    supported_commands = PbFieldInt('discovery.domain_resp.supported_commands')


from whad.rf4ce.stack.apl.profiles.mso.parsers import PeripheralIDsValue,\
    VersioningValue, BatteryStatusValue, ValidationConfigurationValue
from struct import pack

class InformationAttribute:
    def __init__(self, name, identifier, value, writable=False):
        self.name = name
        self.identifier = identifier
        self.writable = writable
        self.value = value


class InformationBase:
    def __init__(self):
        self.attributes = [
            InformationAttribute(
                name="Peripheral IDs",
                identifier=0x00,
                value=PeripheralIDsValue.pack([]),
                writable=True
            ),
            InformationAttribute(
                name="RF Statistics",
                identifier=0x01,
                value=b"\x00"*16,
                writable=True
            ),
            InformationAttribute(
                name="Versioning",
                identifier=0x02,
                value=VersioningValue.pack(
                    (1, 0, 0, 0),
                    (1, 0, 0, 0),
                    (1, 0, 0, 0),
                ),
                writable=True
            ),
            InformationAttribute(
                name="Battery Status",
                identifier=0x03,
                value=BatteryStatusValue.pack(
                    (False, False, False),
                    4.0,
                    0,
                    0,
                    4.0
                ),
                writable=True
            ),
            InformationAttribute(
                name="Short RF Retry Period",
                identifier=0x04,
                value=pack("I", 100000),
                writable=False
            ),
            InformationAttribute(
                name="IR-RF Database",
                identifier=0xDB,
                value=b"",
                writable=False
            ),
            InformationAttribute(
                name="Validation Configuration",
                identifier=0xDC,
                value=ValidationConfigurationValue.pack(61680,61680),
                writable=False
            ),
            InformationAttribute(
                name="General Purpose",
                identifier=0xFF,
                value = b"\x00" * 16 * 32,
                writable=True
            ),
        ]

    def get(self, filter = lambda _ : True):
        for attribute in self.attributes:
            if filter(attribute):
                return attribute
        return None

    def set(self, filter = lambda _ : True, value=b""):
        for attribute in self.attributes:
            if filter(attribute):
                attribute.value = value
                return attribute
        return None



"""WHAD Protocol Hub

The protocol hub is a set of wrappers for Protobuf messages that provides a way
to support different versions of our communication protocol. This protocol hub
provides a default message parser handling different protocol versions that will
pick the correct message wrapper class to parse it. Message wrappers simplifies
the way protocol buffers messages are created by mapping some of their properties
to protobuf messages fields.
"""
import logging
from typing import Union, Type, Optional

# Load default Python's `StrEnum` class for Python >= 3.11
# and rely on `StrEnum` package to provide this class for
# previous Python versions.
try:
    from enum import StrEnum
except ImportError:
    from strenum import StrEnum


from google.protobuf.message import DecodeError
from scapy.config import conf

from whad.protocol.whad_pb2 import Message
from .registry import Registry

logger = logging.getLogger(__name__)

class Domain(StrEnum):
    """Supported protocols."""

    BLE = 'ble'
    DOT15D4 = 'dot15d4'
    ESB = 'esb'
    PHY = 'phy'
    RF4CE = 'rf4ce'
    UNIFYING = 'unifying'
    ZIGBEE = 'zigbee'

class ProtocolHub(Registry):
    """WHAD Protocol Hub class

    This class is an interface between all our Python code and the devices, that
    support all the existing versions of the WHAD protocol and handles every
    differences that exist between them in a transparent fashion.
    """

    NAME = 'hub'
    LAST_VERSION = 2
    VERSIONS = {}

    @staticmethod
    def set_domain(domain: Domain):
        """Configure the hub for a specific domain.

        This is where we can handle any specific configuration of Scapy
        layers if required.

        :param  domain: Domain to use
        :type   domain: Domain
        """
        # Specify the chosen dot15d4 protocol, if specified
        if domain == Domain.RF4CE:
            conf.dot15d4_protocol = 'rf4ce'
        elif domain == Domain.ZIGBEE:
            conf.dot15d4_protocol = 'zigbee'

    def __init__(self, proto_version: int = LAST_VERSION):
        """Instantiate a WHAD protocol hub for a specific version.
        """
        self.__version = proto_version
        self.__cache = {}

    @property
    def version(self) -> int:
        return self.__version

    @property
    def generic(self):
        return self.get('generic')

    @property
    def discovery(self):
        return self.get('discovery')

    @property
    def ble(self):
        return self.get('ble')

    @property
    def dot15d4(self):
        return self.get('dot15d4')

    @property
    def phy(self):
        return self.get('phy')

    @property
    def esb(self):
        return self.get('esb')

    @property
    def unifying(self):
        return self.get('unifying')

    def load(self, factory: str) -> Optional[Type[Registry]]:
        """Lazy loading of factory class (dedicated registry).

        :param factory: Factory alias to load
        :type  factory: str
        :return: Factory class
        :rtype: Registry
        """
        if factory == 'generic':
            from .generic import Generic
            return Generic
        elif factory == 'discovery':
            from .discovery import Discovery
            return Discovery
        elif factory == 'ble':
            from .ble import BleDomain
            return BleDomain
        elif factory in ('dot15d4', 'rf4ce', 'zigbee'):
            from .dot15d4 import Dot15d4Domain
            return Dot15d4Domain
        elif factory == 'esb':
            from .esb import EsbDomain
            return EsbDomain
        elif factory == 'phy':
            from .phy import PhyDomain
            return PhyDomain
        elif factory == 'unifying':
            from .unifying import UnifyingDomain
            return UnifyingDomain

        # Not found
        logger.warning(
            "[protohub::load()] cannot dynamically load message definitions for domain %s !",
            factory
        )
        return None

    def get(self, factory: str) -> Registry:
        """
        Dynamically load registry bound to a specific domain
        and find the corresponding factory class.

        :param factory: factory name to load.
        :type  factory: str
        """
        if factory not in self.__cache:
            self.__cache[factory] = self.load(factory)
        return ProtocolHub.bound(factory, self.__version)(self.__version)

    def parse(self, data: Union[Message, bytes]):
        """Parse a serialized WHAD message into an associated object.
        """
        if isinstance(data, bytes):
            try:
                # Use protocol buffers to parse our message
                msg = Message()
                msg.ParseFromString(bytes(data))
            except DecodeError:
                # Error occurred when parsing message
                logger.debug("Decoding error occurred when parsing %s", data)
                return None
        elif isinstance(data, Message):
            msg = data
        else:
            return None

        # Identify the message base type and ensure the corresponding
        # factory has been loaded.
        msg_type = msg.WhichOneof('msg')
        if msg_type not in self.__cache:
            self.__cache[msg_type] = self.load(msg_type)

        # Process incoming message, forward to corresponding sub-registries
        # following the defined registry hierarchy.
        return ProtocolHub.bound(
            msg.WhichOneof('msg'),
            self.__version).parse(self.__version, msg)

    def convert_packet(self, packet):
        """Convert packet to the corresponding message.
        """
        msg = None
        # We dispatch packets based on their layers
        if self.ble.is_packet_compat(packet):
            logger.debug('[hub] convert_packet(): packet is BLE')
            msg = self.ble.convert_packet(packet)
        elif self.dot15d4.is_packet_compat(packet):
            logger.debug('[hub] convert_packet(): packet is Dot15d4')
            msg = self.dot15d4.convert_packet(packet)
        elif self.esb.is_packet_compat(packet):
            logger.debug('[hub] convert_packet(): packet is ESB')
            msg = self.esb.convert_packet(packet)
        elif self.phy.is_packet_compat(packet):
            logger.debug('[hub] convert_packet(): packet is PHY')
            msg = self.phy.convert_packet(packet)
        elif self.unifying.is_packet_compat(packet):
            logger.debug('[hub] convert_packet(): packet is Unifying')
            msg = self.unifying.convert_packet(packet)
        else:
            logger.error('[hub] convert_packet(): packet is unknown !')

        return msg




"""Zigbee utility for WHAD

This utility is exposed in WHAD as `zigbee-enddevice` and provides the following
features:

- Network scanning
- Network association
- Device emulation

$ zigbee-enddevice scan -> find the first usable BLE-capable device
$ zigbee-enddevice -i uart scan
$ zigbee-enddevice associate <network pan id>
$ zigbee-enddevice -> spawns a shell

"""
from whad.cli.app import CommandLineApp, run_app

from .helpers import create_enddevice
from .shell import ZigbeeEndDeviceShell
from .commands import *


class ZigbeeEndDeviceApp(CommandLineApp):

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD ZigBee End device utility',
            interface=True,
            commands=True
        )

        self.add_argument(
            '--network-panid',
            '-t',
            dest='network_panid',
            help='Specify target network extended Pan ID'
        )

        self.add_argument(
            '--file',
            '-f',
            dest='script',
            help='Specify a script to run'
        )


    def run(self):
        """Override App's run() method to handle scripting feature.
        """
        # Launch pre-run tasks
        self.pre_run()

        if self.args.script is not None:
            if self.is_piped_interface():
                # Make sure we have all the required parameters
                failed = False
                for param in []:
                    if not hasattr(self.args, param):
                        self.error('Source interface does not provide a Zigbee connection')
                        failed = True
                        break

                if not failed:
                    # Create central device
                    enddevice, _ = create_enddevice(self, piped=True)

                    if enddevice is not None:
                        myshell = ZigbeeEndDeviceShell(
                            self.input_interface,
                            connector=enddevice,
                            extended_address=self.args.network_panid
                        )
                        myshell.run_script(self.args.script)
                    else:
                        self.error('Failed to open piped interface.')

            # We need to have an interface specified
            elif self.interface is not None:
                # Launch an interactive shell (well, driven by our script)
                myshell = ZigbeeEndDeviceShell(self.interface)
                myshell.run_script(self.args.script)
            else:
                self.error('You need to specify an interface with option --interface.')
        else:
            super().run(pre=False, post=False)

        # Launch post-run tasks
        self.post_run()

def zigbee_enddevice_main():
    app = ZigbeeEndDeviceApp()
    run_app(app)



"""ESB Packet translator
"""
import logging

from scapy.packet import Packet

from whad.scapy.layers.esb import ESB_Hdr, ESB_Payload_Hdr, ESB_Ack_Response, ESB_Pseudo_Packet

from whad.hub.esb import generate_esb_metadata
from whad.hub import ProtocolHub
from whad.hub.message import HubMessage
from whad.hub.esb import RawPduReceived as EsbRawPduReceived, PduReceived as EsbPduReceived
from whad.hub.unifying import RawPduReceived as UniRawPduReceived, PduReceived as UniPduReceived


logger = logging.getLogger(__name__)

class ESBMessageTranslator:
    """ESB Whad message translator.

    This translator is used to provide the format of a specific scapy packet
    as well as standard methods to convert WHAD ESB messages into scapy packets
    (if it makes sense) and scapy packets into WHAD ESB messages.
    """

    def __init__(self, domain="esb", protocol_hub: ProtocolHub=None):
        self.__cached_address = None
        self.__domain = domain
        self.__hub = protocol_hub

    def format(self, packet):
        """
        Converts a scapy packet with its metadata to a tuple containing a scapy packet with
        the appropriate header and the timestamp in microseconds.
        """
        if ESB_Hdr not in packet:
            packet = ESB_Hdr(address=self.__cached_address)/packet

        packet.preamble = 0xAA # force a rebuild
        formatted_packet = ESB_Pseudo_Packet(bytes(packet))

        timestamp = None
        if hasattr(packet, "metadata"):
            timestamp = packet.metadata.timestamp

        return formatted_packet, timestamp


    def from_message(self, message: HubMessage) -> Packet:
        """Translate a specific WHAD message into the corresponding scapy packet.

        :param message: WHAD message
        :return: Corresponding scapy packet
        :rtype: Packet
        """
        try:
            if isinstance(message, (EsbRawPduReceived, UniRawPduReceived)):
                packet = ESB_Hdr(bytes(message.pdu))
                packet.preamble = 0xAA # force a rebuild

                if ESB_Payload_Hdr not in packet:
                    packet = packet/ESB_Payload_Hdr()/ESB_Ack_Response()

                packet.metadata = generate_esb_metadata(message)
                return packet

            if isinstance(message, (EsbPduReceived, UniPduReceived)):
                packet = ESB_Payload_Hdr(bytes(message.pdu))
                packet.metadata = generate_esb_metadata(message)
                return packet

            # Cannot convert message into packet
            return None
        except AttributeError:
            return None

    def from_packet(self, packet, channel=None, retransmission_count=1) -> HubMessage:
        """Translate a packet into the corresponding WHAD message.

        :param packet: Packet to translate
        :type packet: Packet
        :param channel: Channel to use
        :type channel: int, optional
        :param retransmission_count: Maximum number of retransmissions
        :type retransmission_count: int
        :return: Hub message
        :rtype: HubMessage
        """
        if ESB_Hdr in packet:
            # Force packet preamble to 0xAA
            packet.preamble = 0xAA

            # Create a SendRawPdu message
            msg = self.__hub.get(self.__domain).create_send_raw_pdu(
                channel if channel is not None else 0xFF,
                bytes(packet),
                retransmission_count
            )
        elif ESB_Payload_Hdr in packet:
            # Create a SendPdu message
            msg = self.__hub.get(self.__domain).create_send_pdu(
                channel if channel is not None else 0xFF,
                bytes(packet),
                retransmission_count
            )
        else:
            msg = None
        return msg



"""BLE characteristic read command handler
"""
import logging
from hexdump import hexdump

from prompt_toolkit import print_formatted_text, HTML
from whad.hub.ble.bdaddr import BDAddress
from whad.ble.profile.attribute import UUID, InvalidUUIDException
from whad.cli.app import command
from whad.ble.stack.att.exceptions import AttError
from whad.ble.stack.gatt.exceptions import GattTimeoutException
from whad.ble.cli.central.helpers import show_att_error, create_central

# Global logger
logger = logging.getLogger(__name__)

# Expected parameters that must be passed to our program
# to use an already established connection
EXPECTED_BLE_PARAMS = [
    "initiator_bdaddr",
    "initiator_addrtype",
    "target_bdaddr",
    "target_addrtype",
    "conn_handle",
]

def read_gatt_characteristic(app, command_args, device, profile_loaded=False):
    """Read the value of a GATT characteristic based on the given command arguments.

    :param command_args: arguments to pass to 'read'
    :type command_args: list
    :param device: Peripheral device from which the characteristic value has to be read
    :type device: :class:`whad.ble.profile.PeripheralDevice`
    :param profile_loaded: pre-loaded GATT profile
    :type profile_loaded: GattProfile 
    """
    # parse target arguments
    if len(command_args) == 0:
        app.error('You must provide at least a characteristic value handle or characteristic UUID.')
        return

    handle = None
    offset = None

    # figure out what the handle is
    if command_args[0].lower().startswith('0x'):
        try:
            handle = int(command_args[0].lower(), 16)
        except ValueError:
            app.error("Wrong handle: '%s'", command_args[0])
            return
    else:
        try:
            handle = int(command_args[0])
        except ValueError:
            try:
                handle = UUID(command_args[0].replace('-',''))
            except InvalidUUIDException:
                app.error("Wrong UUID: '%s'", command_args[0])
                return

    # Check offset and length
    if len(command_args) >= 2:
        try:
            offset = int(command_args[1])
        except ValueError:
            app.error("Wrong offset value, will use 0 instead.")
            offset = None

    # Perform characteristic read by handle
    if not isinstance(handle, UUID):
        try:
            value = device.read(handle, offset=offset)

            # Display result as hexdump
            if len(value) > 0:
                hexdump(value)
            else:
                print_formatted_text(HTML('<i>Empty data</i>'))

        except AttError as att_err:
            show_att_error(app, att_err)
        except GattTimeoutException:
            app.error("GATT timeout while reading.")

    else:
        if not profile_loaded:
            # Perform discovery if UUID is given
            device.discover()

        # Search characteristic from its UUID
        target_charac = device.find_characteristic_by_uuid(handle)
        if target_charac is not None:
            try:
                # Read data
                if offset is not None:
                    value = target_charac.read(offset=offset)
                else:
                    value = target_charac.read()

                # Display result as hexdump
                if len(value) > 0:
                    hexdump(value)
                else:
                    print_formatted_text(HTML('<i>Empty data</i>'))

            except AttError as att_err:
                show_att_error(app, att_err)
            except GattTimeoutException:
                app.error("GATT timeout while reading.")
        else:
            app.error("No characteristic found with UUID %s", handle)

@command('read')
def read_handler(app, command_args):
    """read a GATT attribute
    
    <ansicyan><b>read</b> <i>[UUID | handle] ([offset])</i></ansicyan>

    Read an attribute identified by its handle, or read the value of a characteristic
    identified by its UUID (if unique). An optional offset can be provided
    to start reading from the specified byte position (it will issue a
    <i>ReadBlob</i> operation).

    Result is displayed as an hexadecimal dump with corresponding ASCII text:

    $ whad-ble -i hci0 -b 00:11:22:33:44:55 read 42
    00000000: 74 68 69 73 20 69 73 20  61 20 74 65 73 74        this is a test
    """
    # We need to have an interface specified
    if app.interface is not None and app.args.bdaddr is not None:

        # Make sure BD address is valid
        if not BDAddress.check(app.args.bdaddr):
            app.error("Invalid BD address: %s", app.args.bdaddr)
            return

        # Create Central connector based on app configuration
        central, profile_loaded = create_central(app, piped=False)

        # If no connector returned, there was an error, simply exit.
        if central is None:
            return

        # Start central mode
        central.start()

        # Connect to target device
        device = central.connect(app.args.bdaddr, random=app.args.random)
        if device is None:
            app.error("Cannot connect to %s, device does not respond.", app.args.bdaddr)
        else:
            # Read GATT characteristic
            read_gatt_characteristic(app, command_args, device, profile_loaded)

            # Disconnect
            device.disconnect()

        # Terminate central
        central.stop()

    # Piped interface
    elif app.args.bdaddr is None and app.is_piped_interface():

        # Make sure we have all the required parameters
        for param in EXPECTED_BLE_PARAMS:
            if not hasattr(app.args, param):
                app.error("Source interface does not provide a BLE connection")

        # Create Central connector based on app configuration
        central, profile_loaded = create_central(app, piped=True)

        # If no connector returned, there was an error, simply exit.
        if central is None:
            return

        # Retrieve connected device
        device = central.peripheral()

        # Read GATT characteristic
        read_gatt_characteristic(app, command_args, device, profile_loaded)

        # Disconnect
        device.disconnect()
        central.stop()
        central.close()

    elif app.interface is not None:
        app.error("You need to specify a target device with option --bdaddr.")



from dataclasses import dataclass, field
from whad.hub.phy import Endianness

@dataclass
class FSKConfiguration:
    """
    Configuration for FSK modulation.

    :param deviation: select the modulation deviation (dev)
    """
    deviation : int = 500000

@dataclass
class LoRaConfiguration:
    """
    Configuration for LoRa modulation.

    :param spreading_factor: select the spreading factor (sf)
    :param coding_rate: select the coding rate (cr)
    :param bandwidth: select the bandwidth (bw)
    :param enable_crc: enable LoRa CRC (crc)
    :param enable_explicit_mode: enable LoRa explicit mode (em)
    :param preamble_length: set LoRa preamble length in symbols (pl)
    """
    spreading_factor : int = 7
    coding_rate : int = 45
    bandwidth : int = 125000
    enable_crc: bool = False
    enable_explicit_mode: bool = False
    preamble_length: int = 12


@dataclass
class InjectionConfiguration:
    """
    Configuration for injecting Phy communications.

    :param frequency: select the frequency to sniff (f)
    :param little_endian: select little endianness (le)
    :param datarate: select number of bits per second (d)
    :param packet_size: select packet size (s)
    :param sync_word: select synchronization word (w)
    :param ask: select ASK modulation (ask)
    :param gfsk: select GFSK modulation (gfsk)
    :param bfsk: select BFSK modulation (bfsk)
    :param qfsk: select QFSK modulation (qfsk)
    :param bpsk: select BPSK modulation (bpsk)
    :param qpsk: select QPSK modulation (qpsk)
    :param lora: select LoRa modulation (lora)
    """
    frequency : int = None
    little_endian : bool = False
    datarate : int = None
    packet_size : int = None
    sync_word : bytes = None
    ask: bool = False
    gfsk: bool = False
    bfsk: bool = False
    qfsk: bool = False
    bpsk: bool = False
    qpsk: bool = False
    lora: bool = False
    fsk_configuration: FSKConfiguration = None
    lora_configuration: LoRaConfiguration = None



"""WHAD Enhanced ShockBurst protocol

This module exposes a dedicated connector for the ESB
protocol, as well as a sniffer, a scanner, a receiver (PRX)
and a transmitter (PTX) connectors.
"""

from .connector import ESB, Sniffer, PRX, PTX, Scanner
from .utils.phy import PHYS

__all__ = [
    'ESB',
    'Sniffer',
    'Scanner',
    'PRX',
    'PTX',
    'PHYS'
]



from whad.common.stack import Layer

class Dot15d4Manager(Layer):
    """
    Defines a 802.15.4 manager, containing some services.
    """
    def configure(self, options={}):
        super().configure(options=options)
        self._services = {}
        self.init()

    def init(self):
        pass

    def add_service(self, service_name, service):
        self._services[service_name] = service

    def get_service(self, service):
        if service in self._services:
            return self._services[service]
        return None

    @property
    def upper_layer(self):
        for k, v in self.layers.items():
            if self.get_layer(k).parent.alias == self.alias:
                return self.get_layer(k)
        return None

    @property
    def lower_layer(self):
        return self.parent


    @property
    def database(self):
        return self.state

    @property
    def services(self):
        return self._services



"""WHAD Protocol PHY mode messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from whad.protocol.phy.phy_pb2 import MonitorCmd, StartCmd, StopCmd
from ..message import pb_bind, PbFieldInt, PbFieldBool, PbFieldArray, PbMessageWrapper
from . import PhyDomain

from .timestamp import Timestamp

@pb_bind(PhyDomain, 'sniff', 1)
class SniffMode(PbMessageWrapper):
    """PHY Sniff mode message
    """

    iq_stream = PbFieldBool('phy.sniff.iq_stream', optional=True)

@pb_bind(PhyDomain, 'jam', 1)
class JamMode(PbMessageWrapper):
    """PHY Jam mode message
    """

    mode = PbFieldInt('phy.jam.mode')

@pb_bind(PhyDomain, 'jammed', 1)
class Jammed(PbMessageWrapper):
    """PHY Jammed notification message
    """

    timestamp = PbFieldInt('phy.jammed.timestamp')

@pb_bind(PhyDomain, 'monitor', 1)
class MonitorMode(PbMessageWrapper):
    """PHY Monitor mode message
    """

    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.phy.monitor.CopyFrom(MonitorCmd())

@pb_bind(PhyDomain, 'monitor_report', 1)
class MonitoringReport(PbMessageWrapper):
    """PHY monitoring report notification message
    """

    timestamp = PbFieldInt('phy.monitor_report.timestamp')
    reports = PbFieldArray('phy.monitor_report.report')

@pb_bind(PhyDomain, 'start', 1)
class Start(PbMessageWrapper):
    """PHY Start mode message
    """

    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.phy.start.CopyFrom(StartCmd())

@pb_bind(PhyDomain, 'stop', 1)
class Stop(PbMessageWrapper):
    """PHY Stop mode message
    """

    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.phy.stop.CopyFrom(StopCmd())


import re
from whad.dot15d4.exceptions import InvalidDot15d4AddressException
from struct import pack, unpack

class Dot15d4Address(object):
    """This class represents a 802.15.4 address.
    """

    SHORT = 0x00
    EXTENDED = 0x01

    def __init__(self, address):
        """Initialize 802.15.4 address

        Adapt to input parameters to represent a 802.15.4 short or extended address.
        """

        if isinstance(address, str):
            if re.match(r'^([0-9a-fA-F]{2}\:){7}[0-9a-fA-F]{2}$', address) is not None:
                self.__type = Dot15d4Address.EXTENDED
                self.__value = bytes.fromhex(address.replace(":", ""))
            elif re.match('[0-9a-fA-F]{16}$', address) is not None:
                self.__type = Dot15d4Address.EXTENDED
                self.__value = bytes.fromhex(address)
            elif re.match('[0-9a-fA-F]{4}$', address) is not None:
                self.__type = Dot15d4Address.SHORT
                self.__value = bytes.fromhex(address)
            elif re.match('0x[0-9a-fA-F]{4}$', address) is not None:
                self.__type = Dot15d4Address.SHORT
                self.__value = bytes.fromhex(address[2:])
            else:
                raise InvalidDot15d4AddressException
        elif isinstance(address, int):
            if address <= 0xFFFF:
                self.__type = Dot15d4Address.SHORT
                self.__value = pack('>H', address)
            else:
                self.__type = Dot15d4Address.EXTENDED
                self.__value = pack('>Q', address)
        else:
            raise InvalidDot15d4AddressException

    def __eq__(self, other):
        if not isinstance(other, Dot15d4Address):
            other = Dot15d4Address(other)

        return (self.value == other.value) and (self.__type == other.type)

    def __str__(self):
        if self.__type == Dot15d4Address.EXTENDED:
            return ':'.join(['%02x' % b for b in self.__value])
        else:
            return "0x" + "".join(['%02x' % b for b in self.__value])

    def __repr__(self):
        return 'Dot15d4Address(%s)' % str(self)

    @property
    def type(self):
        return self.__type

    @property
    def value(self):
        if self.__type == Dot15d4Address.EXTENDED:
            return unpack('<Q', self.__value)[0]
        elif self.__type == Dot15d4Address.SHORT:
            return unpack('>H', self.__value)[0]

    def is_extended(self):
        """Determine if address is extended.

        :rtype: bool
        :return: True if 802.15.4 address is extended, False otherwise.
        """
        return (self.__type == Dot15d4Address.EXTENDED)

    @staticmethod
    def from_bytes(addr_bytes):
        """Convert a 2 or 8-byte array into a valid 802.15.4 address.

        :param bytes addr_bytes: Network or Device address as a bytearray.
        :rtype: Dot15d4Address
        :return: An instance of Dot15d4Address representing the corresponding Dot15d4 address.
        """
        hex_address = addr_bytes[::-1].hex()
        if len(addr_bytes) == 8:
            address = ":".join([hex_address[i*2:(i+1)*2] for i in range(int(len(hex_address)/2))])
            return Dot15d4Address(address)
        elif len(addr_bytes) == 2:
            address = "".join([hex_address[i*2:(i+1)*2] for i in range(int(len(hex_address)/2))])
            return Dot15d4Address(address)
        else:
            raise InvalidDot15d4AddressException



"""Generic stack testing features
"""

from whad.common.stack import Layer, alias

class contextual(object):

    def __init__(self, value):
        self.__value = value

    def __call__(self, func):
        func.is_contextual = self.__value
        return func

class LayerMessage(object):
    '''This class is used by `Sandbox` to represent a message between two layers,
    keeping track of the source, destination, data, tag and named arguments.
    '''

    def __init__(self, source, destination, data, tag='default', **kwargs):
        self.__source = source
        self.__destination = destination
        self.__data = data
        self.__tag = tag
        self.__args = kwargs

    @property
    def source(self):
        return self.__source
    
    @property
    def destination(self):
        return self.__destination
    
    @property
    def data(self):
        return self.__data
    
    @property
    def tag(self):
        return self.__tag
    
    @property
    def args(self):
        return self.__args

    def __repr__(self):
        return 'LayerMessage(source="%s", destination="%s", data="%s", tag="%s")' % (
            self.source,
            self.destination,
            bytes(self.data),
            self.tag
        )

    def __eq__(self, other):
        # check main properties
        if (self.source != other.source) or (self.destination != other.destination) \
            or (self.data != other.data) or (self.tag != other.tag):
            return False
        
        # check arguments
        for arg in self.args:
            if arg not in other.args:
                print('LayerMessage - comparison failed: missing arg %s' % arg)
                return False
            if self.args[arg] != other.args[arg]:
                print('LayerMessage - comparison failed: arg %s values are different (%s, %s)' % (
                    arg, self.args[arg], other.args[arg])
                )
                return False
            
        # Same.
        print('same')
        return True
        

@alias('sandbox')
class Sandbox(Layer):
    """This layer is used as a container for one or more layers that need
    to be tested.
    """

    def __init__(self, parent=None, layer_name=None, options={}, target=None):
        super().__init__(parent=parent, layer_name=layer_name, options=options)
        self.messages = []
        #self.target = target.alias
        #self.target_class = target

    def populate(self, options={}):
        '''Populate static layers and install a custom message monitor callback
        '''
        super().populate(options=options)

        # Install a monitor callback on all sub-layers
        for layer in self.layers:
            self.layers[layer].register_monitor_callback(self.log_message)

    def instantiate(self, clazz):
        '''Instantiate a layer and install a custom message monitor callback
        '''
        layer = super().instantiate(clazz)
        layer.register_monitor_callback(self.log_message)
        return layer

    def get_layer(self, name: str):
        '''Retrieve a specific layer if one of ours, or this instance if layer
        cannot be found.
        '''
        layer = super().get_layer(name)
        if layer is None:
            self.destination = self
            return self
        else:
            return layer
    
    def get_handler(self, source, tag):
        '''This method is called by our stack message handler search code, and is
        required here to return our dummy message handler.
        '''
        return self.dummy_message_handler

    def log_message(self, source, destination, data, tag='default', **kwargs):
        '''Layer message monitoring.
        '''
        self.messages.append(LayerMessage(
            source,
            destination,
            data,
            tag=tag,
            **kwargs
        ))

    @contextual(True)
    def dummy_message_handler(self, source, data, **kwargs):
        """Dummy message handler, does nothing.
        """
        pass

    def expect(self, messages, strict=False):
        '''Checks if one or more messages have been captured.
        '''
        if isinstance(messages, list):
            if strict and len(self.messages) != len(messages):
                return False
            if len(self.messages) >= len(messages):
                for i in range(len(self.messages)):
                    if messages[i] != self.messages[i]:
                        return False
                return True
            else:
                return False
        elif isinstance(messages, LayerMessage):
            if len(self.messages) == 0:
                print('no message received')
                return False
            # Check if this message has been seen in our monitored messages
            for msg in self.messages:
                if msg == messages:
                    return True
            return False

    




